"""Bliss-side for Flint ID32 plots"""

from ..flint import BasePlot


class SpecGenPlot(BasePlot):
    WIDGET = "blissoda.id32.widgets.SpecGenPlotWidget"

    def update_plot(
        self, outputfile_sum: str, scan_number: int, max_scans: int
    ) -> None:
        # print(f"{outputfile_sum=}, {scan_number=}, {max_scans=}")
        self.submit("update_plot", outputfile_sum, scan_number, max_scans)



"""
Utilities for the EXAFS scans.

- Continuous scans: `energy scan framework <https://gitlab.esrf.fr/bliss/energy_scan>`_ which
  uses the `fast scan framework <https://gitlab.esrf.fr/bliss/fscan>`_.

- Step scans: `BM23 step energy scan framework <https://gitlab.esrf.fr/bm23/bm23/-/blob/master/bm23/BM23scanExafs.py>`_.
"""

from typing import Union

from ..import_utils import unavailable_type

try:
    from bliss.scanning.scan import Scan as BlissScanType
except ImportError as ex:
    BlissScanType = unavailable_type(ex)

try:
    from fscan.trigscan import TrigScanCustomRunner as TrigScanCustomRunnerType
except ImportError as ex:
    TrigScanCustomRunnerType = unavailable_type(ex)


ScanType = Union[BlissScanType, TrigScanCustomRunnerType]


def is_multi_xas_scan(scan: ScanType) -> bool:
    try:
        # fscan.trigscan.TrigScanPars in case of TrigScanCustomRunnerType
        return scan.pars.nscans > 1
    except AttributeError:
        return False


def multi_xas_subscan_size(scan: ScanType) -> int:
    try:
        # fscan.trigscan.TrigScanPars in case of TrigScanCustomRunnerType
        return scan.pars.npoints + 1

        # "npoints" is a badly chosen variable name.
        # It should be called "nsteps" and `size = nstep + 1`.
    except AttributeError:
        return 0



from __future__ import annotations

import logging
import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

from ..persistent.parameters import ParameterInfo
from ..persistent.parameters import autocomplete_property
from ..processor import BlissScanType
from ..resources import resource_filename
from ..xrpd.processor import XrpdProcessor

MULTI_RADIAL_BINS = "multi_radial_bins"
AZIMUTHAL_RANGES = "azimuthal_ranges"

logger = logging.getLogger(__name__)


class Id22XrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("pyfai_config", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
        ParameterInfo(MULTI_RADIAL_BINS, category="PyFai"),
        ParameterInfo(AZIMUTHAL_RANGES, category="PyFai"),
    ],
):
    DEFAULT_WORKFLOW = resource_filename("id22", "Sum_then_integrate_with_saving.json")
    MULTI_CONFIG_WORKFLOW = resource_filename(
        "id22", "Sum_then_multiintegrate_with_saving.json"
    )
    DEFAULT_WORKFLOW_NO_SAVE = None

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault(
            "integration_options",
            {
                "method": "no_csr_ocl_gpu",
                "nbpt_rad": 4096,
                "unit": "q_nm^-1",
            },
        )
        defaults.setdefault(MULTI_RADIAL_BINS, None)
        defaults.setdefault(AZIMUTHAL_RANGES, None)
        defaults.setdefault("trigger_at", "END")

        super().__init__(config=config, defaults=defaults)

        # Disable data from memory for now
        # The data structure is indeed different when getting it from file or from memory
        self._set_parameter("data_from_memory", False)

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        return self.pyfai_config

    def get_integration_options(
        self, scan: BlissScanType, lima_name: str
    ) -> Optional[dict]:
        integration_options = self.integration_options
        if integration_options:
            return integration_options.to_dict()
        return None

    def get_inputs(self, scan, lima_name: str) -> List[dict]:
        inputs = super().get_inputs(scan, lima_name)
        inputs += self.get_sum_inputs(scan, lima_name)
        inputs += self.get_save_ascii_inputs(scan, lima_name)
        inputs += self.get_multi_tasks_inputs(scan, lima_name)
        return inputs

    def get_sum_inputs(self, scan, lima_name: str):
        task_identifier = "SumBlissScanImages"

        filename = self.get_filename(scan)
        scan_nb = scan.scan_info.get("scan_nb")

        inputs = [
            {
                "task_identifier": task_identifier,
                "name": "filename",
                "value": filename,
            },
            {
                "task_identifier": task_identifier,
                "name": "output_filename",
                "value": self.master_output_filename(scan),
            },
            {
                "task_identifier": task_identifier,
                "name": "scan",
                "value": scan_nb,
            },
            {
                "task_identifier": task_identifier,
                "name": "detector_name",
                "value": lima_name,
            },
            {
                "task_identifier": task_identifier,
                "name": "monitor_name",
                "value": self.monitor_name,
            },
            {
                "task_identifier": task_identifier,
                "name": "flush_period",
                "value": self.flush_period,
            },
        ]
        tscan_info = scan.scan_info.get("tscan_info")
        if tscan_info:
            background_step = tscan_info.get("background_step")
            if background_step is not None:
                inputs.append(
                    {
                        "task_identifier": task_identifier,
                        "name": "background_step",
                        "value": background_step,
                    }
                )
        if self.data_from_memory:
            scan_memory_url = f"{scan.root_node.db_name}:{scan._node_name}"
            inputs.append(
                {
                    "task_identifier": task_identifier,
                    "name": "scan_memory_url",
                    "value": scan_memory_url,
                }
            )
        inputs += self._get_data_access_inputs(task_identifier)
        return inputs

    def get_save_ascii_inputs(self, scan, lima_name):
        filename = self.get_filename(scan)
        scan_nb = scan.scan_info.get("scan_nb")
        root = self.scan_processed_directory(scan)
        stem = os.path.splitext(os.path.basename(filename))[0]
        basename = f"{stem}_{scan_nb}_{lima_name}_integrated.dat"
        return [
            {
                "task_identifier": "SaveAsciiPattern1D",
                "name": "filename",
                "value": os.path.join(root, basename),
            },
        ]

    def get_save_inputs(self, scan, lima_name, task_identifier):
        inputs = super().get_save_inputs(scan, lima_name, task_identifier)
        inputs += [
            {
                "task_identifier": task_identifier,
                "name": "nxprocess_name",
                "value": f"{lima_name}_integrate",
            }
        ]
        return inputs

    def get_multi_tasks_inputs(self, scan, lima_name):
        # Integration
        task_identifier = "MultiConfigIntegrate1D"
        inputs = self.get_integrate_inputs(scan, lima_name, task_identifier)
        multi_configs = self._get_multi_configs()
        if multi_configs:
            inputs.append(
                {
                    "task_identifier": task_identifier,
                    "name": "configs",
                    "value": multi_configs,
                }
            )

        # Saving
        task_identifier = "SaveNexusMultiPattern1D"
        inputs += super().get_save_inputs(scan, lima_name, task_identifier)
        inputs.append(
            {
                "task_identifier": task_identifier,
                "name": "nxprocess_name",
                "value": f"{lima_name}_multi_integrate",
            }
        )

        # ASCII Saving
        if multi_configs is not None:
            filename = self.get_filename(scan)
            scan_nb = scan.scan_info.get("scan_nb")
            root = self.scan_processed_directory(scan)
            stem = os.path.splitext(os.path.basename(filename))[0]
            filenames = []
            for config in multi_configs:
                basename = f"{stem}_{scan_nb}_{lima_name}"
                if "nbpt_rad" in config:
                    basename += f'_{config["nbpt_rad"]}rbins'
                if "azimuth_range_min" in config:
                    basename += f'_az{config["azimuth_range_min"]}-{config["azimuth_range_max"]}'
                basename += "_integrated.dat"
                filenames.append(os.path.join(root, basename))
            inputs.append(
                {
                    "task_identifier": "SaveAsciiMultiPattern1D",
                    "name": "filenames",
                    "value": filenames,
                }
            )
        return inputs

    def _get_multi_configs(self) -> Optional[List[dict]]:
        if self.multi_radial_bins and not self.azimuthal_ranges:
            return [{"nbpt_rad": npt} for npt in self.multi_radial_bins]

        if self.azimuthal_ranges and not self.multi_radial_bins:
            return [
                {"azimuth_range_min": az_min, "azimuth_range_max": az_max}
                for (az_min, az_max) in self.azimuthal_ranges
            ]

        if self.multi_radial_bins and self.azimuthal_ranges:
            configs = []
            for npt in self.multi_radial_bins:
                for az_min, az_max in self.azimuthal_ranges:
                    configs.append(
                        {
                            "nbpt_rad": npt,
                            "azimuth_range_min": az_min,
                            "azimuth_range_max": az_max,
                        }
                    )
            return configs

        return None

    @autocomplete_property
    def multi_radial_bins(self):
        return self._get_parameter(MULTI_RADIAL_BINS)

    @multi_radial_bins.setter
    def multi_radial_bins(self, bins_list: Optional[Sequence[int]]):
        if bins_list is not None and not isinstance(bins_list, Sequence):
            logging.warning(
                "multi_radial_bins must be a list of values. Ex: [100, 200, 400]"
            )
            return
        self._set_parameter(MULTI_RADIAL_BINS, bins_list)
        self._update_workflow()

    @autocomplete_property
    def azimuthal_ranges(self):
        return self._get_parameter(AZIMUTHAL_RANGES)

    @azimuthal_ranges.setter
    def azimuthal_ranges(self, ranges: Optional[Sequence[Tuple[float, float]]]):
        if ranges is not None and not isinstance(ranges, Sequence):
            logging.warning(
                "azimuthal_ranges must be a list of 2-size sequences. Ex: [(-100, 100), (45, 65), (135, 155)]"
            )
            return
        self._set_parameter(AZIMUTHAL_RANGES, ranges)
        self._update_workflow()

    def _update_workflow(self):
        if self.multi_radial_bins is None and self.azimuthal_ranges is None:
            self.workflow_with_saving = self.DEFAULT_WORKFLOW
        else:
            self.workflow_with_saving = resource_filename(
                "id22", "Sum_then_multiintegrate_with_saving.json"
            )

    def _data_to_plot_url(self, scan, lima_name):
        if not scan.scan_info.get("save"):
            return None

        output_url = self.online_output_url(scan, lima_name)
        if self.multi_radial_bins is None and self.azimuthal_ranges is None:
            return f"{output_url}/{lima_name}_integrate/integrated"
        else:
            return f"{output_url}/{lima_name}_multi_integrate_0/integrated"



from .import_utils import UnavailableObject

# Proxy to the current Bliss session object (if any).
try:
    from bliss import current_session
except ImportError as ex:
    current_session = UnavailableObject(ex)
else:
    try:
        current_session.name
    except AttributeError as ex:
        current_session = UnavailableObject(ex)

# Proxy to a namespace used by Bliss scripts to access objects
# and functions from the current Bliss session (if any).
try:
    from bliss import setup_globals
except ImportError as ex:
    setup_globals = UnavailableObject(ex)



from blissoda.bliss_globals import setup_globals
from blissoda.demo.id09 import txs_processor


def id09_txs_test(expo=0.2, npoints=10):
    try:
        txs_processor.enable()
        pct(expo)
        setup_globals.loopscan(
            npoints,
            expo,
            setup_globals.difflab6,
            setup_globals.diode1,
            setup_globals.diode2,
        )
    finally:
        txs_processor.disable()


def pct(expo):
    s = setup_globals.ct(
        expo,
        setup_globals.difflab6,
        setup_globals.diode1,
        setup_globals.diode2,
    )
    txs_processor.trigger_workflow_on_new_scan(s)
    return s



import logging
import os
from typing import Any
from typing import Dict
from typing import Optional

from ewoksjob.client import get_future
from ewoksjob.client import submit

from ..bliss_globals import current_session
from ..persistent.parameters import ParameterInfo
from ..processor import BaseProcessor
from ..processor import BlissScanType
from ..utils.directories import get_dataset_processed_dir
from .plotter import Id02Plotter

logger = logging.getLogger(__name__)


class Id02BaseProcessor(
    BaseProcessor,
    parameters=[
        ParameterInfo("queue", category="workflows", deprecated_names=["worker"]),
        ParameterInfo("number_of_scans", category="plotting"),
    ],
    deprecated_class_attributes={"DEFAULT_WORKER": "DEFAULT_QUEUE"},
):
    DEFAULT_QUEUE = None

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("queue", self.DEFAULT_QUEUE)
        defaults.setdefault("number_of_scans", 4)

        super().__init__(config=config, defaults=defaults)

        self._preset = self._set_up_preset()

        self._plotter = Id02Plotter(number_of_scans=self.number_of_scans)
        self._plotter.replot()

    def _set_up_preset(self):
        raise NotImplementedError()

    def scan_requires_processing(self, scan: BlissScanType) -> bool:
        return scan.scan_info["save"]

    def get_workflow(self, scan: BlissScanType) -> dict:
        dets = [d.name for d in self._preset.getDetectors(scan)]
        return self._preset.buildWorkflow(dets)

    def get_inputs(self, scan: BlissScanType) -> list:
        dets = [d.name for d in self._preset.getDetectors(scan)]
        return self._preset.getInputs(scan, dets)

    def get_filename(self, scan: BlissScanType) -> str:
        filename = scan.scan_info.get("filename")
        if filename:
            return filename
        return current_session.scan_saving.filename

    def scan_processed_directory(self, scan: BlissScanType) -> str:
        return get_dataset_processed_dir(self.get_filename(scan))

    def workflow_destination(self, scan: BlissScanType) -> str:
        filename = self.get_filename(scan)
        scan_nb = scan.scan_info.get("scan_nb")
        root = self.scan_processed_directory(scan)
        stem = os.path.splitext(os.path.basename(filename))[0]
        basename = f"{stem}_{scan_nb:04d}.json"
        return os.path.join(root, basename)

    def _trigger_workflow_on_new_scan(self, scan: BlissScanType) -> Optional[Any]:
        if not self.scan_requires_processing(scan):
            return None

        workflow = self.get_workflow(scan)
        kwargs = {"inputs": self.get_inputs(scan), "outputs": [{"all": False}]}
        if scan.scan_info.get("save"):
            kwargs["convert_destination"] = self.workflow_destination(scan)

        future = submit(args=(workflow,), kwargs=kwargs, queue=self.queue)

        # TODO: Handle plotting with Flint
        future = get_future(future.uuid)

        return future



from __future__ import annotations

import functools
import logging
import time
from typing import Dict
from typing import Optional
from typing import Tuple

import numpy
from silx.utils.retry import RetryTimeoutError

from ..bliss_globals import current_session
from ..flint.plotter import BasePlotter
from .models import XrpdFieldName
from .models import XrpdPlotInfo
from .plot_data import XrpdNXDataInfo
from .plot_data import create_plot_info
from .plot_data import delete_old_entries
from .plot_data import delete_plot_info
from .plot_data import get_plots
from .plot_data import get_plots_to_remove
from .plot_data import get_xrpd_nxdata_info
from .plots import Xrpd2dIntegrationPlot
from .plots import XrpdCurvePlot
from .plots import XrpdImagePlot
from .utils import axis_label

logger = logging.getLogger(__name__)


def _background_task(method):
    @functools.wraps(method)
    def wrapper(*args, **kw):
        try:
            return method(*args, **kw)
        except Exception as e:
            logging.error("XRPD plotter task failed (%s)", e, exc_info=True)
            raise

    return wrapper


class XrpdPlotter(BasePlotter):
    def __init__(self, number_of_scans: int = 0) -> None:
        super().__init__(number_of_scans)
        if self._HAS_BLISS:
            nb_removed_entries = delete_old_entries(current_session)
            if nb_removed_entries:
                logging.info(f"Removed {nb_removed_entries} old entries from Redis")

    def handle_workflow_result(
        self,
        future,
        lima_name: str,
        scan_name: str,
        output_url: Optional[str] = None,
        retry_timeout: int = 60,
        retry_period: int = 1,
    ):
        """Handle workflow results in a background task"""
        if output_url:
            func = self._handle_workflow_result_from_file
        else:
            func = self._handle_workflow_result_from_memory
        self._spawn(
            func,
            future,
            scan_name,
            lima_name,
            output_url=output_url,
            retry_timeout=retry_timeout,
            retry_period=retry_period,
        )

    def replot(self, **retry_options) -> None:
        """Re-draw all plots."""
        for plot_info in get_plots():
            self._spawn(self._update_plot, plot_info, retry_options)

    def clear(self):
        """Clear all plots."""
        plot_infos = get_plots()
        lima_names = set()
        for plot_info in plot_infos:
            lima_names.add(plot_info.lima_name)
            self._remove_plot(plot_info)
        for lima_name in lima_names:
            self.clear_lima_plots(lima_name)

    @_background_task
    def _handle_workflow_result_from_memory(
        self,
        future,
        scan_name: str,
        lima_name: str,
        retry_timeout: Optional[int] = 600,
        **retry_options,
    ) -> None:
        retry_period = retry_options.get("retry_period")
        result = future.result(timeout=retry_timeout, interval=retry_period)

        if result["radial_units"] is None:
            return

        radial_name, radial_unit = result["radial_units"].split("_")
        radial_label = axis_label(radial_name, radial_unit)

        plot_data: Dict[XrpdFieldName, numpy.ndarray] = {
            "radial": result["radial"],
            "intensity": result["intensity"],
        }

        if result.get("azimuthal") is not None:
            plot_data["azimuthal"] = result["azimuthal"]
            azim_name, azim_unit = result["azimuthal_units"].split("_")
            azim_label = axis_label(azim_name, azim_unit)
        else:
            azim_label = None

        plot_info = create_plot_info(
            scan_name,
            lima_name,
            radial_label=radial_label,
            azim_label=azim_label,
            plot_data=plot_data,
        )

        self._remove_old_plots()
        self._update_plot(plot_info, retry_options)

    @_background_task
    def _handle_workflow_result_from_file(
        self,
        future,
        scan_name: str,
        lima_name: str,
        output_url: str,
        **retry_options,
    ) -> None:
        retry_period = retry_options.get("retry_period")

        nxdata_info = self._load_nxdata_info(future, output_url, retry_period)
        if nxdata_info is None:
            return

        plot_info = create_plot_info(
            scan_name,
            lima_name,
            nxdata_info.radial_label,
            nxdata_info.azim_label,
            plot_data=nxdata_info.plot_data,
            hdf5_url=nxdata_info.intensity_url,
        )
        self._remove_old_plots()

        while not future.done():
            self._update_plot(plot_info, retry_options)
            if retry_period:
                time.sleep(retry_period)
        self._update_plot(plot_info, retry_options)

    def _load_nxdata_info(
        self,
        future,
        nxdata_url: str,
        retry_period: float,
    ) -> Optional[XrpdNXDataInfo]:
        """Try to read from nxdata_url until the future is ready"""
        while not future.done():
            try:
                return get_xrpd_nxdata_info(
                    nxdata_url,
                    # Use retry_period as a short timeout to check the future
                    retry_timeout=retry_period,
                    retry_period=retry_period,
                )
            except RetryTimeoutError:
                pass

        if future.exception():
            return None

        # Give it a last chance for 10 seconds
        try:
            return get_xrpd_nxdata_info(
                nxdata_url,
                retry_timeout=10,
                retry_period=retry_period,
            )
        except RetryTimeoutError:
            return None

    def _update_plot(self, plot_info: XrpdPlotInfo, retry_options: dict) -> None:
        for plot in self._get_plots(plot_info):
            plot.update_plot(plot_info.pk, retry_options)

    def _remove_plot(self, plot_info: XrpdPlotInfo) -> None:
        for plot in self._get_plots(plot_info):
            plot.remove_plot(plot_info.pk)
        delete_plot_info(plot_info)

    def _remove_old_plots(self):
        plot_infos = get_plots_to_remove(self._max_plots)
        for plot_info in plot_infos:
            self._remove_plot(plot_info)

    def _get_plots(
        self, plot_info: XrpdPlotInfo
    ) -> Tuple[Xrpd2dIntegrationPlot] | Tuple[XrpdCurvePlot, XrpdImagePlot]:
        if "azimuthal" in plot_info.field_names:
            last_2d = super()._get_plot(
                f"2D Integrated {plot_info.lima_name} (last)", Xrpd2dIntegrationPlot
            )
            return (last_2d,)
        accum_1d = super()._get_plot(f"Integrated {plot_info.lima_name}", XrpdImagePlot)
        last_1d = super()._get_plot("Integrated (Last)", XrpdCurvePlot)
        return (last_1d, accum_1d)

    def clear_lima_plots(self, lima_name: str) -> None:
        last_1d = super()._get_plot("Integrated (Last)", XrpdCurvePlot)
        last_1d.clear()
        accum_1d = super()._get_plot(f"Integrated {lima_name}", XrpdImagePlot)
        accum_1d.clear()
        last_2d = super()._get_plot(
            f"2D Integrated {lima_name} (last)", Xrpd2dIntegrationPlot
        )
        last_2d.clear()



"""Flint-side for Flint XRPD plots"""

import logging
from typing import List

from silx.gui.plot import Plot1D
from silx.gui.plot import Plot2D

from ..flint import capture_errors
from .models import XrpdPlotInfo
from .plot_data import append_image_data
from .plot_data import get_2d_integration_data
from .plot_data import get_curve_data

logger = logging.getLogger(__name__)


class XrpdCurveWidget(Plot1D):
    @capture_errors
    def remove_plot(self, plot_key: str) -> None:
        try:
            plot_info = XrpdPlotInfo.get(plot_key)
        except KeyError as ex:
            logger.debug("%r already removed: %s", plot_key, ex)
            return

        legend = plot_info.legend
        logger.debug("remove %r", legend)
        self.remove(legend=plot_info.legend)

    @capture_errors
    def update_plot(self, plot_key: str, retry_options: dict) -> None:
        x, y, plot_info = get_curve_data(plot_key, **retry_options)
        if y is None:
            return
        legend = plot_info.legend
        if len(x) != len(y):
            logger.error("XRPD curve plot %r: Nx=%d, Ny=%d", legend, len(x), len(y))
        else:
            logger.debug("XRPD curve plot %r: Nx=%d, Ny=%d", legend, len(x), len(y))
            self.addCurve(
                x, y, legend=legend, xlabel=plot_info.radial_label, ylabel="Intensity"
            )

    @capture_errors
    def get_labels(self) -> List[str]:
        return [legend.getName() for legend in self.getItems()]


class XrpdImageWidget(Plot2D):
    @capture_errors
    def remove_plot(self, plot_key: str) -> None:
        self.clear()

    @capture_errors
    def update_plot(self, plot_key: str, retry_options: dict) -> None:
        img = self.getImage(legend=plot_key)
        if img is None:
            current_data = None
        else:
            current_data = img.getData(copy=False)
        x, y, plot_info = append_image_data(plot_key, current_data, **retry_options)
        if y is None:
            return
        origin = x[0], 0
        scale = x[1] - x[0], 1
        self.clear()
        title = plot_info.legend
        self.setGraphTitle(title)
        logger.debug("XRPD image plot %s: %s points", title, len(y))
        self.addImage(
            y,
            legend=title,
            xlabel=plot_info.radial_label,
            ylabel="Scan points",
            origin=origin,
            scale=scale,
        )

    @capture_errors
    def get_labels(self) -> List[str]:
        return [self.getGraphTitle()]


class Xrpd2dIntegrationWidget(Plot2D):
    @capture_errors
    def remove_plot(self, plot_key: str) -> None:
        self.clear()

    @capture_errors
    def update_plot(self, plot_key: str, retry_options: dict) -> None:
        x, y, intensity, plot_info = get_2d_integration_data(plot_key, **retry_options)
        if x is None or y is None or intensity is None:
            return
        origin = x[0], y[0]
        scale = x[1] - x[0], y[1] - y[0]
        self.clear()
        title = plot_info.legend
        self.setGraphTitle(title)
        logger.debug(
            "XRPD 2D integration plot %r: Nx=%d, Ny=%d, Shape=%s",
            title,
            len(x),
            len(y),
            intensity.shape,
        )
        self.addImage(
            intensity,
            legend=title,
            xlabel=plot_info.radial_label,
            ylabel=plot_info.azim_label,
            origin=origin,
            scale=scale,
        )

    @capture_errors
    def get_labels(self) -> List[str]:
        return [self.getGraphTitle()]



import logging
import os
import shutil
from glob import glob
from typing import Any
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional

from ewoksjob.client import submit

from ..bliss_globals import current_session
from ..flint.access import WithFlintAccess
from ..import_utils import unavailable_module
from ..persistent.parameters import ParameterInfo
from ..processor import BaseProcessor
from ..processor import BlissScanType
from ..resources import resource_filename
from ..utils import directories
from .plots import TemperaturePlot

try:
    import gevent
except ImportError as ex:
    gevent = unavailable_module(ex)


logger = logging.getLogger(__name__)


class WorkflowType(NamedTuple):
    xas: bool
    fit: bool

    def parameter_name(self) -> str:
        name = "workflow"
        if self.fit:
            name += "_with_fit"
        if self.xas:
            name = "xas_" + name
        return name


class Id24TemperaturePlotter(
    BaseProcessor,
    parameters=[
        ParameterInfo("workflow", category="workflows"),
        ParameterInfo("workflow_with_fit", category="workflows"),
        ParameterInfo("xas_workflow", category="workflows"),
        ParameterInfo("xas_workflow_with_fit", category="workflows"),
        ParameterInfo("energy_name", category="XAS"),
        ParameterInfo("mu_name", category="XAS"),
        ParameterInfo("extend_plotrange_left", category="plot", doc="(nm)"),
        ParameterInfo("extend_plotrange_right", category="plot", doc="(nm)"),
        ParameterInfo("two_color_difference", category="plot", doc="(nm)"),
        ParameterInfo("dpi", category="plot"),
        ParameterInfo("refit", category="fit"),
        ParameterInfo("wavelength_min", category="fit", doc="(nm)"),
        ParameterInfo("wavelength_max", category="fit", doc="(nm)"),
    ],
):
    DEFAULT_WORKFLOWS = {
        WorkflowType(xas=True, fit=True): resource_filename(
            "id24", "id24_xas_planck_fitplot.json"
        ),
        WorkflowType(xas=True, fit=False): resource_filename(
            "id24", "id24_xas_planck_plot.json"
        ),
        WorkflowType(xas=False, fit=True): resource_filename(
            "id24", "id24_planck_fitplot.json"
        ),
        WorkflowType(xas=False, fit=False): resource_filename(
            "id24", "id24_planck_plot.json"
        ),
    }

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("trigger_at", "END")

        for workflow_type, workflow_filename in self.DEFAULT_WORKFLOWS.items():
            defaults.setdefault(workflow_type.parameter_name(), workflow_filename)

        defaults.setdefault("energy_name", "energy_enc")
        defaults.setdefault("mu_name", "mu_trans")

        defaults.setdefault("extend_plotrange_left", -15)
        defaults.setdefault("extend_plotrange_right", 50)
        defaults.setdefault("two_color_difference", 42)
        defaults.setdefault("dpi", 150)

        defaults.setdefault("refit", False)
        defaults.setdefault("wavelength_min", None)
        defaults.setdefault("wavelength_max", None)

        super().__init__(config=config, defaults=defaults)

        self.plotter = Plotter()

    def _info_categories(self) -> Dict[str, dict]:
        categories = super()._info_categories()
        ntasks = self.plotter.purge_tasks()
        status = categories.setdefault("status", {})
        status["Fitting"] = self.refit
        status["Plotting tasks"] = ntasks
        return categories

    def _trigger_workflow_on_new_scan(self, scan: BlissScanType) -> Optional[dict]:
        if not self._scan_requires_workflow(scan):
            return
        filename = scan.scan_info["filename"]
        scan_number = scan.scan_info["scan_nb"]
        workflow_type = self.get_workflow_type(scan)
        self._trigger_workflow(workflow_type, scan_number, filename=filename)

    def get_workflow_type(self, scan: BlissScanType) -> WorkflowType:
        return WorkflowType(xas=self._is_xas_scan(scan), fit=self.refit)

    def trigger_workflow(
        self,
        scan_number: int,
        filename: Optional[str] = None,
        xas: bool = True,
    ) -> None:
        self._trigger_workflow(
            WorkflowType(xas=xas, fit=self.refit), scan_number, filename=filename
        )

    def _trigger_workflow(
        self,
        workflow_type: WorkflowType,
        scan_number: int,
        filename: Optional[str] = None,
    ) -> None:
        if filename:
            if not os.path.isabs(filename):
                basename = os.path.basename(filename)
                raw_data = directories.get_raw_dir(current_session.scan_saving.filename)
                filenames = glob(os.path.join(raw_data, "*", "*", basename))
                if not filenames:
                    raise FileNotFoundError(filename)
                filename = filenames[0]
        else:
            filename = current_session.scan_saving.filename
        workflow = self._get_workflow(workflow_type, filename)

        future = submit(
            args=(workflow,),
            kwargs={
                "inputs": self._get_workflow_inputs(
                    workflow_type, filename, scan_number
                ),
                "outputs": [{"all": False}],
            },
        )
        self.plotter.handle_workflow_result(future)

    def _scan_requires_workflow(self, scan: BlissScanType) -> bool:
        if not self._has_temperature(scan):
            return False
        if not scan.scan_info.get("save"):
            return False
        if not scan.scan_info.get("filename"):
            return False
        if not scan.scan_info.get("scan_nb"):
            return False
        return True

    def _has_temperature(self, scan: BlissScanType) -> bool:
        channels = scan.scan_info.get("channels", dict())
        return "laser_heating_down:T_planck" in channels

    def _is_xas_scan(self, scan: BlissScanType) -> bool:
        channels = scan.scan_info.get("channels", dict())
        energy_suffix = f":{self.energy_name}"
        return any(name.endswith(energy_suffix) for name in channels)

    def _get_workflow(
        self, workflow_type: WorkflowType, filename: str
    ) -> Optional[str]:
        """Get the workflow to execute for the scan and ensure it is located
        in the proposal directory for user reference and worker accessibility.
        """
        parameter_name = workflow_type.parameter_name()
        src_file = self._get_parameter(parameter_name)
        if src_file is None:
            return
        if not os.path.isfile(src_file):
            src_file = self.DEFAULT_WORKFLOWS[workflow_type]

        workflow_directory = directories.get_workflows_dir(filename)
        dst_file = os.path.join(workflow_directory, os.path.basename(src_file))
        if src_file != dst_file:
            self._set_parameter(parameter_name, dst_file)

        if not os.path.exists(dst_file):
            os.makedirs(workflow_directory, exist_ok=True)
            shutil.copyfile(src_file, dst_file)

        return dst_file

    def _get_workflow_inputs(
        self, workflow_type: WorkflowType, filename: str, scan_number: int
    ) -> List[dict]:
        inputs = self._get_read_inputs(workflow_type, filename, scan_number)
        inputs += self._get_plot_inputs(workflow_type, filename)
        inputs += self._get_fit_inputs()
        return inputs

    def _get_read_inputs(
        self, workflow_type: WorkflowType, filename: str, scan_number: int
    ) -> List[dict]:
        if workflow_type.xas:
            task_identifier = "XasTemperatureRead"
            subscan_number = 2
        else:
            task_identifier = "ScanTemperatureRead"
            subscan_number = 1
        return [
            {
                "task_identifier": task_identifier,
                "name": "filename",
                "value": filename,
            },
            {
                "task_identifier": task_identifier,
                "name": "scan_number",
                "value": scan_number,
            },
            {
                "task_identifier": task_identifier,
                "name": "subscan_number",
                "value": subscan_number,
            },
            {
                "task_identifier": task_identifier,
                "name": "retry_timeout",
                "value": 60,
            },
            {
                "task_identifier": task_identifier,
                "name": "energy_name",
                "value": self.energy_name,
            },
            {
                "task_identifier": task_identifier,
                "name": "mu_name",
                "value": self.mu_name,
            },
        ]

    def _get_plot_inputs(
        self, workflow_type: WorkflowType, filename: str
    ) -> List[dict]:
        if workflow_type.xas:
            task_identifier = "XasTemperaturePlot"
        else:
            task_identifier = "ScanTemperaturePlot"
        output_directory = directories.get_processed_subdir(filename, "temperature")
        return [
            {
                "task_identifier": task_identifier,
                "name": "output_directory",
                "value": output_directory,
            },
            {
                "task_identifier": task_identifier,
                "name": "extend_plotrange_left",
                "value": self.extend_plotrange_left,
            },
            {
                "task_identifier": task_identifier,
                "name": "extend_plotrange_right",
                "value": self.extend_plotrange_right,
            },
            {
                "task_identifier": task_identifier,
                "name": "two_color_difference",
                "value": self.two_color_difference,
            },
            {
                "task_identifier": task_identifier,
                "name": "dpi",
                "value": self.dpi,
            },
        ]

    def _get_fit_inputs(self) -> List[dict]:
        if not self.refit:
            return []
        task_identifier = "PlanckRadianceFit"
        return [
            {
                "task_identifier": task_identifier,
                "name": "wavelength_min",
                "value": self.wavelength_min,
            },
            {
                "task_identifier": task_identifier,
                "name": "wavelength_max",
                "value": self.wavelength_max,
            },
        ]


class Plotter(WithFlintAccess):
    def __init__(self) -> None:
        super().__init__()
        self._tasks = list()

    def purge_tasks(self) -> int:
        """Remove references to tasks that have finished."""
        self._tasks = [t for t in self._tasks if t]
        return len(self._tasks)

    def kill_tasks(self) -> int:
        """Kill all tasks."""
        gevent.killall(self._tasks)
        return self.purge_tasks()

    def handle_workflow_result(self, *args, **kwargs) -> None:
        """Handle workflow results in a background task"""
        self._spawn(self._handle_workflow_result, *args, **kwargs)

    def _handle_workflow_result(self, future, timeout: int = 60):
        try:
            results = future.result(timeout=timeout)
            if not results:
                return
            filenames = results["filenames"]
            directory = os.path.dirname(filenames[0])

            plot = self._get_plot()
            plot.select_directory(directory)
        except Exception as e:
            logger.exception(e)

    def _spawn(self, *args, **kw):
        task = gevent.spawn(*args, **kw)
        self._tasks.append(task)
        self.purge_tasks()

    def _get_plot(self) -> TemperaturePlot:
        return super()._get_plot("Temperature", TemperaturePlot)



import os
import shutil
import sys
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Generator

if sys.version_info < (3, 9):
    import importlib_resources
else:
    import importlib.resources as importlib_resources


@contextmanager
def resource_path(*args) -> Generator[Path, None, None]:
    """The resource is specified relative to `blissoda.resources`.

    .. code-block:: python

        with resource_path("exafs", "exafs.ows") a path:
            ...
    """
    source = importlib_resources.files(__name__).joinpath(*args)
    with importlib_resources.as_file(source) as path:
        if not path.is_file():
            raise FileNotFoundError(f"Not a blissoda resource file: '{path}'")
        yield path


def resource_filename(*args) -> str:
    """The resource is specified relative to `blissoda.resources`.

    .. code-block:: python

        filename = resource_filename("exafs", "exafs.ows")
    """
    with resource_path(*args) as path:
        pass
    if not path.exists():
        # resource was extract from zip: copy for persistency
        with resource_path(*args) as path:
            path_copy = Path(tempfile.mkdtemp()) / path.name
            shutil.copyfile(str(path), str(path_copy))
            path = path_copy
    return str(path)


def copy_resource_glob(*args: str, dest_dir: Path = Path(".")) -> None:
    r"""
    Copy all resource files matching the given glob pattern to the destination directory.

    :param args: A glob pattern, e.g. ``('xrpd', '*.json')`` or ``('exafs', '*.ows')``.
    :param dest_dir: The directory where the files will be copied (default: current directory).
    """
    dest_dir = Path(dest_dir)
    dest_dir.mkdir(parents=True, exist_ok=True)

    resources_root = importlib_resources.files(__name__)

    for resource in resources_root.rglob(os.path.join(*args)):
        if resource.is_file():
            with importlib_resources.as_file(resource) as src_path:
                shutil.copy(src_path, dest_dir / src_path.name)



"""Automatic pyfai integration for every scan with saving and plotting"""

from typing import Any
from typing import Dict
from typing import Optional

from ..persistent.parameters import ParameterInfo
from ..processor import BlissScanType
from ..xrpd.processor import XrpdProcessor


class ID24XrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("pyfai_config", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
    ],
):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:

        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault(
            "integration_options",
            {
                "method": "no_csr_ocl_gpu",
                "nbpt_rad": 4096,
                "unit": "q_nm^-1",
            },
        )

        super().__init__(config=config, defaults=defaults)

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        return self.pyfai_config

    def get_integration_options(
        self, scan: BlissScanType, lima_name: str
    ) -> Optional[dict]:
        integration_options = self.integration_options
        if integration_options:
            return integration_options.to_dict()
        return None



import pytest

from .. import resources


def test_existing():
    with resources.resource_path("exafs", "exafs.ows") as path:
        assert path.is_file()

    assert str(path) == resources.resource_filename("exafs", "exafs.ows")


def test_non_existing():
    with pytest.raises(FileNotFoundError):
        with resources.resource_path("exafs", "notexisting.ows"):
            pass

    with pytest.raises(FileNotFoundError):
        resources.resource_filename("exafs", "notexisting.ows")


def test_not_a_file():
    with pytest.raises(FileNotFoundError):
        with resources.resource_path("exafs"):
            pass

    with pytest.raises(FileNotFoundError):
        resources.resource_filename("exafs")


def test_copy_existing(tmp_path):
    resources.copy_resource_glob("xrpd", "*.json", dest_dir=tmp_path)
    copied_files = {f.name for f in tmp_path.iterdir() if f.is_file()}
    expected = {"integrate_scan_without_saving.json", "integrate_scan_with_saving.json"}
    assert copied_files == expected


def test_copy_non_existing(tmp_path):
    resources.copy_resource_glob("xrpd", "*.notexisting", dest_dir=tmp_path)
    copied_files = {f.name for f in tmp_path.iterdir() if f.is_file()}
    expected = set()
    assert copied_files == expected



REAL_DATA_ROOT = "/data/visitor"

DEMO_DATA_ROOT = "/data/scisoft/ewoks"

DEFAULT_OUT_DIRNAME = "ewoks_results"

# Fast scanning: not all motors have units
AXES_UNITS = {
    "id21_sxm": {
        "samy": "mm",
        "samz": "mm",
        "sampy": "um",
        "sampz": "um",
    },
    "id21_nano": {
        "nsy": "mm",
        "nsz": "mm",
        "nspy": "um",
        "nspz": "um",
    },
    "id16b": {
        "samy": "mm",
        "samz": "mm",
        "sampy": "um",
        "sampz": "um",
    },
}

# Mosaic XRF mapping
VIRTUAL_AXES = {
    "id21_sxm": {
        "sy": "<samy>+<sampy>",
        "sz": "<samz>+<sampz>",
    },
    "id21_nano": {
        "sy": "<nsy>+<nspy>",
        "sz": "<nsz>+<nspz>",
    },
    "id16b": {
        "sy": "<samy>+<sampy>",
        "sz": "<samz>+<sampz>",
    },
}

IGNORE_AXES = {
    "id21_sxm": [],
    "id21_nano": ["nsz1", "nsz2", "nsz3"],
    "id16b": [],
}

ENERGY_COUNTER = {
    "id21_sxm": "Edcm",
    "id21_nano": "Edcm",
    "id16b": "Edcm",
}

I0_COUNTER = {
    "id21_sxm": "iodet",
    "id21_nano": "niodet",
    "id16b": None,
}

MCA_NAME_FORMAT = {
    "id21_sxm": "fx_sxm_det{}",
    "id21_nano": "fx_nano_det{}",
    "id16b": "fxb_det{}",
}

DEFAULT_LIVETIME_REF_VALUE = "0.1"

DEFAULT_COUNTER_REF_VALUE = "np.nanmean(<instrument/{}/data>)"

DO_PROFILE = False



from typing import List
from typing import Union

import numpy

from ...import_utils import unavailable_class

try:
    from blissdata.data.events.channel import ChannelDataEvent
    from blissdata.streaming import DataStream
except ImportError as ex:
    ChannelDataEvent = unavailable_class(ex)
    DataStream = unavailable_class(ex)


class PersistentNdArray:
    def __init__(self, redis_key: str) -> None:
        self._datastream = DataStream(redis_key)

    def append(self, value: numpy.ndarray) -> None:
        self._datastream.add_event(_encode(value[numpy.newaxis, ...]))

    def extend(self, value: numpy.ndarray) -> None:
        self._datastream.add_event(_encode(value))

    def __getitem__(self, idx) -> Union[numpy.ndarray, List[numpy.ndarray]]:
        if idx == 0:
            events = self._datastream.range(count=1)
            if not events:
                raise IndexError("index out of range")
            adict = events[0][1]
            return _decode(adict)[0]
        if idx == -1:
            events = self._datastream.rev_range(count=1)
            if not events:
                raise IndexError("index out of range")
            adict = events[0][1]
            return _decode(adict)[-1]

        events = self._datastream.range()
        if events:
            event = ChannelDataEvent.merge(events)
            arr = _data_from_event(event)
        else:
            arr = numpy.array([])
        if idx == ():
            return arr
        return arr[idx]

    def remove(self) -> None:
        self._datastream.clear()


def _encode(data: numpy.ndarray) -> ChannelDataEvent:
    desc = {"shape": data.shape[1:], "dtype": data.dtype}
    return ChannelDataEvent(data, desc)


def _decode(data: dict) -> numpy.ndarray:
    event = ChannelDataEvent(raw=data)
    return _data_from_event(event)


def _data_from_event(event: ChannelDataEvent) -> numpy.ndarray:
    if event.npoints == 1:
        return event.data[numpy.newaxis, ...]
    return event.data



from __future__ import annotations

from pathlib import Path
from typing import Literal
from typing import Optional
from typing import Union

from pydantic import BaseModel
from pydantic import Field
from pydantic import field_validator

from ..resources import resource_filename


class TomoProcessorModel(BaseModel, validate_assignment=True):
    workflow: str = Field(
        default="tomo_processor.json",
        description="Blissoda resources workflow file",
        examples=["tomo_processor.json"],
    )
    queue: Optional[str] = Field(
        default=None,
        description="Ewoks queue to submit the workflow to",
        examples=["tomo_queue"],
    )
    nabu_config_file: Optional[str] = None
    slice_index: Union[Literal["first", "middle", "last"], int] = "middle"
    rotation_axis_position: Union[
        int,
        float,
        Literal[
            "centered",
            "sliding-window",
            "global",
            "growing-window",
            "sino-coarse-to-fine",
            "composite-coarse-to-fine",
            "octave-accurate",
        ],
    ] = "sliding-window"
    phase_retrieval_method: Union[Literal["CTF", "Paganin", "None"], None] = "None"
    delta_beta: Union[str, float, int] = "100"
    show_last_slice: bool = Field(
        default=False,
        description="Display last reconstructed slice in Flint",
    )

    @field_validator("workflow")
    def check_workflow(cls, v):
        if not isinstance(v, str):
            raise ValueError("workflow must be a string")
        if not v.endswith(".json"):
            raise ValueError("workflow must be a JSON file")
        if not Path(resource_filename("tomo", v)).exists():
            raise FileNotFoundError(f"Workflow file {v} does not exist")
        return v

    @field_validator("queue")
    def check_queue(cls, v):
        if v is not None and not isinstance(v, str):
            raise ValueError("queue must be a string or None")
        return v

    @field_validator("nabu_config_file")
    def check_nabu_config_file(cls, v):
        if v is not None and not isinstance(v, str):
            raise ValueError("nabu_config_file must be a string or None")
        if v and not Path(v).exists():
            raise FileNotFoundError(f"Nabu config file {v} does not exist")
        return v

    @field_validator("slice_index")
    def check_slice_index(cls, v):
        if isinstance(v, str):
            if v not in ["first", "last", "middle"]:
                raise ValueError(
                    "slice_index must be 'first', 'last', 'middle' or an integer"
                )
        elif not isinstance(v, int):
            raise ValueError(
                "slice_index must be an integer or one of 'first', 'last', 'middle'"
            )
        return str(v)

    @field_validator("rotation_axis_position")
    def check_rotation_axis_position(cls, v):
        valid_positions = [
            "centered",
            "sliding-window",
            "global",
            "growing-window",
            "sino-coarse-to-fine",
            "composite-coarse-to-fine",
            "octave-accurate",
        ]
        if isinstance(v, str):
            if v not in valid_positions:
                raise ValueError(
                    f"rotation_axis_position must be one of {valid_positions}"
                )
        elif not isinstance(v, (int, float)):
            raise ValueError(
                f"rotation_axis_position must be a float, int, or one of {valid_positions}"
            )
        return v

    @field_validator("phase_retrieval_method")
    def check_phase_method(cls, v):
        if v is None:
            return "None"
        elif v not in ["CTF", "Paganin", "None"]:
            raise ValueError("phase_retrieval_method must be 'CTF', 'Paganin' or None")
        return v

    @field_validator("delta_beta")
    def check_positive_delta_beta(cls, v):
        if float(v) <= 0:
            raise ValueError("delta_beta must be positive")
        return str(v)



"""Bliss-side for Flint XRPD plots"""

from typing import List

from ..flint import BasePlot


class _XrpdBasePlot(BasePlot):
    def remove_plot(self, plot_key: str) -> None:
        self.submit("remove_plot", plot_key)

    def update_plot(self, plot_key: str, hdf5_options: dict) -> None:
        self.submit("update_plot", plot_key, hdf5_options)

    def get_labels(self) -> List[str]:
        return self.submit("get_labels")

    def clear(self) -> None:
        return self.submit("clear")


class XrpdCurvePlot(_XrpdBasePlot):
    WIDGET = "blissoda.xrpd.widgets.XrpdCurveWidget"


class XrpdImagePlot(_XrpdBasePlot):
    WIDGET = "blissoda.xrpd.widgets.XrpdImageWidget"


class Xrpd2dIntegrationPlot(_XrpdBasePlot):
    WIDGET = "blissoda.xrpd.widgets.Xrpd2dIntegrationWidget"



from contextlib import contextmanager
from typing import Any
from typing import Dict
from typing import Generator
from typing import Optional

import fabio
import numpy
from ewoksutils.task_utils import task_inputs

from ..bliss_globals import setup_globals
from ..persistent.parameters import ParameterInfo
from ..streamline.scanner import StreamlineScanner
from ..utils import validators
from . import optimize_exposure
from .utils import ensure_shutter_open


class Id31StreamlineScanner(
    StreamlineScanner,
    parameters=[
        ParameterInfo(
            "optimize_pixel_value", category="exposure/attenuator", validator=float
        ),
        ParameterInfo(
            "optimize_nb_frames", category="exposure/attenuator", validator=int
        ),
        ParameterInfo(
            "optimize_max_exposure_time",
            category="exposure/attenuator",
            validator=float,
        ),
        ParameterInfo("default_attenuator", category="exposure/attenuator"),
        ParameterInfo("attenuator_name", category="names"),
        ParameterInfo("newflat", category="Flat-field", validator=validators.is_file),
        ParameterInfo("oldflat", category="Flat-field", validator=validators.is_file),
        ParameterInfo("flat_enabled", category="Flat-field", validator=bool),
        ParameterInfo("optimize_exposure_per", category="robust vs. speed"),
        ParameterInfo("rockit_distance", category="sample changer"),
        ParameterInfo(
            "optimize_attenuator",
            category="exposure/attenuator",
            validator=bool,
        ),
        ParameterInfo("optimize_mask_file", category="exposure/attenuator"),
    ],
):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ):
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("workflow", "streamline_without_calib_with_flat.json")
        defaults.setdefault("detector_name", "p3")
        defaults.setdefault("attenuator_name", "atten")
        defaults.setdefault("sample_changer_name", "streamline_sc")
        defaults.setdefault(
            "integration_options",
            {
                "method": "no_csr_ocl_gpu",
                "integrator_name": "sigma_clip_ng",
                "extra_options": {"max_iter": 3, "thres": 0},
                "error_model": "azimuthal",  # hybrid gives weird results
                "nbpt_rad": 4096,
                "unit": "q_nm^-1",
            },
        )
        defaults.setdefault("flat_enabled", True)
        defaults.setdefault("newflat", "/data/id31/inhouse/P3/flats.mat")
        defaults.setdefault("oldflat", "/data/id31/inhouse/P3/flats_old.mat")
        defaults.setdefault("optimize_pixel_value", 1e5)
        defaults.setdefault("optimize_nb_frames", 3)
        defaults.setdefault("optimize_max_exposure_time", 4)
        defaults.setdefault("optimize_exposure_per", "baguette")
        defaults.setdefault("rockit_distance", 0.07)
        defaults.setdefault("optimize_attenuator", True)
        defaults.setdefault("optimize_mask_file", None)

        super().__init__(config=config, defaults=defaults)

        self._exposure_conditions = list()
        self._fixed_attenuator_position = None
        self._optimize_mask_array: Optional[numpy.ndarray] = None
        self._update_optimize_mask(self.optimize_mask_file)

    @property
    def optimize_exposure_per(self) -> Optional[str]:
        return self._get_parameter("optimize_exposure_per")

    @optimize_exposure_per.setter
    def optimize_exposure_per(self, value: Optional[str]):
        if value not in (None, "sample", "baguette"):
            raise ValueError("Allowed values are 'sample', 'baguette' or None")
        self._set_parameter("optimize_exposure_per", value)

    @property
    def optimize_mask_file(self) -> Optional[str]:
        return self._get_parameter("optimize_mask_file")

    @optimize_mask_file.setter
    def optimize_mask_file(self, filename: Optional[str]):
        self._update_optimize_mask(filename)

    def _update_optimize_mask(self, filename: Optional[str]):
        if not filename:
            self._optimize_mask_array = None
            self._set_parameter("optimize_mask_file", None)
            return

        try:
            self._optimize_mask_array = fabio.open(filename).data
        except Exception:
            print(f"Error: cannot load optimize_mask_file {filename}, reseting it!")
            self._optimize_mask_array = None
            self._set_parameter("optimize_mask_file", None)
            return

        self._set_parameter("optimize_mask_file", filename)

    @contextmanager
    def run_context(self):
        setup_globals.shopen(
            check_pilatus=False
        )  # check_pilatus = False when the detector was just started
        with super().run_context():
            yield

    def load(self):
        super().load()
        if self.optimize_exposure_per == "baguette":
            self.determine_exposure_conditions()

    def measure_sample(
        self, count_time: float = 1, *args, has_qrcode: bool = True, **kwargs
    ):
        with rockit(self.sample_changer.translation, self.rockit_distance):
            with self._optimize_sample_exposure(
                count_time, has_qrcode=has_qrcode
            ) as expo_time:
                if has_qrcode:
                    expo_time_max = self.optimize_max_exposure_time
                else:
                    expo_time_max = count_time
                expo_time = min(expo_time, expo_time_max)

                if not self.dryrun:
                    ensure_shutter_open()
                try:
                    return super().measure_sample(expo_time, *args, **kwargs)
                except RuntimeError as e:
                    if "Pilatus protection" in str(e):
                        print(
                            f"Skip because of measurement error: {e}. Open shutter again"
                        )
                        return
                    raise

    def determine_exposure_conditions(self):
        """Pre-define optimal conditions: ascan at fixed attenuator position"""
        detector = getattr(setup_globals, self.detector_name)
        if self.default_attenuator is None:
            attenuator = getattr(setup_globals, self.attenuator_name)
            self.default_attenuator = attenuator.bits
        else:
            setup_globals.att(self.default_attenuator)
        self._exposure_conditions = optimize_exposure.optimal_exposure_conditions(
            *self.sample_changer.ascan_arguments(),
            detector,
            tframe=0.2,
            desired_counts=self.optimize_pixel_value,
            nframes_measure=1,
            nframes_default=self.optimize_nb_frames,
            reduce_desired_deviation=True,
            expose_with_integral_frames=False,
        )

    def determine_exposure_conditions_individually(self):
        """Pre-define optimal conditions: ct on each sample with adapted attenuator
        if the default attenuator position gives too much or too little counts"""
        detector = getattr(setup_globals, self.detector_name)
        attenuator = getattr(setup_globals, self.attenuator_name)
        att_value = attenuator.bits
        exposure_conditions = list()
        try:
            for _ in self.sample_changer.iterate_samples_without_qr():
                exposure_conditions.append(self._optimize_exposure_condition(detector))
        finally:
            setup_globals.att(att_value)
        self._exposure_conditions = exposure_conditions

    @contextmanager
    def _optimize_sample_exposure(
        self, count_time: float, has_qrcode: bool = True
    ) -> Generator[float, None, None]:
        """Selecting the optimal measurement conditions and returning the corresponding
        exposure time for the current sample."""

        if not self.optimize_exposure_per:
            # Optimization is disabled
            yield count_time

        elif self.optimize_exposure_per == "baguette":
            # Select pre-defined optimization
            count_time = self._set_exposure_condition()
            yield count_time

        elif not has_qrcode:
            # No QR-code probably means no sample so do not waste time optimizing
            yield count_time

        else:
            # Optimize condition for this sample individually
            detector = getattr(setup_globals, self.detector_name)
            attenuator = getattr(setup_globals, self.attenuator_name)
            att_value = attenuator.bits
            try:
                condition = self._optimize_exposure_condition(detector)
                yield condition.expo_time
            finally:
                setup_globals.att(att_value)

    def _set_exposure_condition(self) -> float:
        if not self._exposure_conditions:
            self.determine_exposure_conditions()
        sample_index = self.sample_changer.current_sample_index
        condition = self._exposure_conditions[sample_index]
        print(f"Pre-defined optimal exposure conditions: {condition}")
        setup_globals.att(condition.att_position)
        return condition.expo_time

    def _optimize_exposure_condition(
        self, detector
    ) -> optimize_exposure.ExposureCondition:
        return optimize_exposure.optimize_exposure_condition(
            detector,
            tframe=0.2,
            default_att_position=self.default_attenuator,
            desired_counts=self.optimize_pixel_value,
            dynamic_range=1 << 20,
            min_counts_per_frame=0,  # take 100
            nframes_measure=1,
            nframes_default=self.optimize_nb_frames,
            reduce_desired_deviation=True,
            expose_with_integral_frames=False,
            optimize_attenuator=self.optimize_attenuator,
            mask=self._optimize_mask_array,
        )

    def init_workflow(self, with_autocalibration: bool = False) -> None:
        if with_autocalibration:
            self.workflow = "streamline_with_calib_with_flat.json"
        else:
            self.workflow = "streamline_without_calib_with_flat.json"

        print(f"Active data processing workflow: {self.workflow}")

    def _job_arguments(self, scan_info, processed_metadata: dict):
        args, kwargs = super()._job_arguments(scan_info, processed_metadata)
        detector_name = self.detector_name

        kwargs["inputs"] += task_inputs(
            task_identifier="FlatFieldFromEnergy",
            inputs={
                "newflat": self.newflat,
                "oldflat": self.oldflat,
                "energy": getattr(setup_globals, self.energy_name).position,
                "enabled": self.flat_enabled and detector_name == "p3",
            },
        )

        # Rely on StreamlineScanner to set url and bliss_scan_url as for other SaveNexusPattern1D tasks
        kwargs["inputs"] += task_inputs(
            task_identifier="SaveNexusPattern1D",
            label="save_q_no_sigmaclip_hdf5",
            inputs={
                "nxprocess_name": f"{detector_name}_integrate_q_no_sigmaclip",
                "nxmeasurement_name": f"{detector_name}_integrated_q_no_sigmaclip",
                "metadata": {
                    f"{detector_name}_integrate_q_no_sigmaclip": {
                        "configuration": {"workflow": self.workflow}
                    }
                },
            },
        )

        # Use workflows from ewoksid31 module
        kwargs["load_options"] = {"root_module": "ewoksid31.workflows"}

        return args, kwargs


@contextmanager
def rockit(motor, distance):
    if distance:
        try:
            with setup_globals.rockit(motor, distance):
                print("ROCKING ON")
                yield
        finally:
            print("ROCKING OFF")
    else:
        print("ROCKING DISABLED")
        yield



from ..import_utils import unavailable_class
from .processor import Id02BaseProcessor

try:
    from id02.acquisition.xpcs_preset import ID02DynamixProcessingPreset
except ImportError as ex:
    ID02DynamixProcessingPreset = unavailable_class(ex)


class Id02XpcsProcessor(Id02BaseProcessor):
    DEFAULT_WORKER = "xpcs"

    def _set_up_preset(self):
        return ID02DynamixProcessingPreset()



from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Dict
from typing import List

from marshmallow import Schema
from marshmallow import fields


@dataclass
class ExposedParameter:
    parameter: str
    title: str
    field_type: fields.Field
    field_options: Dict[str, Any] = field(default_factory=dict)
    field_args: List[Any] = field(default_factory=list)
    read_only: bool = False


class DaiquiriProcessorMixin:
    EXPOSED_PARAMETERS: List[ExposedParameter] = []

    @property
    def parameters_schema(self) -> Schema:
        uiorder = ["state", "state_ok", "enabled"]
        uischema = {
            "enabled": {
                "ui:widget": "BoolButton",
            }
        }

        parameters = {}
        for parameter in self.EXPOSED_PARAMETERS:
            parameters[parameter.parameter] = parameter.field_type(
                metadata={"title": parameter.title, "readOnly": parameter.read_only},
                *parameter.field_args,
                **parameter.field_options,
            )
            uiorder.append(parameter.parameter)

        return type(
            "ExposedParameters",
            (Schema,),
            {
                "Meta": type(
                    "Meta", (object,), {"uiorder": uiorder, "uischema": uischema}
                ),
                "enabled": fields.Bool(metadata={"title": "Enabled"}),
                **parameters,
            },
        )



import sys

import pytest

from ..fluo.parameters.offline import fluoxas_parameters
from ..fluo.parameters.offline import mosaic_xrfmap_parameters
from ..fluo.parameters.offline import xrfmap_parameters


@pytest.mark.skipif(
    sys.platform == "win32", reason="Skipping fluo path tests on windows"
)
def test_xrfmap_single_detector():
    workflow_path, inputs, convert_destination = xrfmap_parameters(
        session="/data/visitor/blc16198/id16b/20250527",
        sample="test",
        dataset="9147_136783",
        scan_number=30,
        config_filenames=["fit.cfg"],
        detector_numbers=[8],
        instrument_name="id16b",
        demo=False,
    )
    eworkflow_path = "/data/visitor/blc16198/id16b/20250527/SCRIPTS/ewoks_results/workflows/xrfmap_single_detector.ows"
    econvert_destination = "/data/visitor/blc16198/id16b/20250527/PROCESSED_DATA/ewoks_results/test_9147_136783_scan0030.json"
    einputs = [
        {
            "name": "filename",
            "value": "/data/visitor/blc16198/id16b/20250527/RAW_DATA/test/test_9147_136783/test_9147_136783.h5",
            "task_identifier": "PickScan",
        },
        {"name": "scan_number", "value": 30, "task_identifier": "PickScan"},
        {
            "name": "output_root_uri",
            "value": "/data/visitor/blc16198/id16b/20250527/PROCESSED_DATA/ewoks_results/test_9147_136783.h5::/30.1",
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "config",
            "value": "/data/visitor/blc16198/id16b/20250527/SCRIPTS/pymca/fit.cfg",
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "fast_fitting",
            "value": True,
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "quantification",
            "value": True,
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "diagnostics",
            "value": True,
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "detector_name",
            "value": "fxb_det8",
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "energy_name",
            "value": "Edcm",
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "counter_name",
            "value": None,
            "task_identifier": "NormalizeXrfResults",
        },
        {
            "name": "counter_normalization_template",
            "value": None,
            "task_identifier": "NormalizeXrfResults",
        },
        {
            "name": "detector_normalization_template",
            "value": "0.1/<instrument/{}/live_time>",
            "task_identifier": "NormalizeXrfResults",
        },
        {
            "name": "axes_units",
            "value": {"samy": "mm", "samz": "mm", "sampy": "um", "sampz": "um"},
            "task_identifier": "RegridXrfResults",
        },
        {
            "name": "ignore_positioners",
            "task_identifier": "RegridXrfResults",
            "value": [],
        },
        {"name": "resolution", "value": None, "task_identifier": "RegridXrfResults"},
    ]

    assert workflow_path == eworkflow_path
    assert convert_destination == econvert_destination
    assert inputs == einputs


@pytest.mark.skipif(
    sys.platform == "win32", reason="Skipping fluo path tests on windows"
)
def test_xrfmap_sum_spectra():
    workflow_path, inputs, convert_destination = xrfmap_parameters(
        session="/data/visitor/ls3288/id21/20240221",
        sample="DIV15_10uM_2",
        dataset="roi75698_90925",
        scan_number=1,
        config_filenames=["sample_10200ev_sheldon.cfg"],
        detector_numbers=[0, 1, 2, 3, 4],
        instrument_name="id21_nano",
        demo=False,
    )
    eworkflow_path = "/data/visitor/ls3288/id21/20240221/SCRIPTS/ewoks_results/workflows/xrfmap_multi_detector_sumspectra.ows"
    econvert_destination = "/data/visitor/ls3288/id21/20240221/PROCESSED_DATA/ewoks_results/DIV15_10uM_2_roi75698_90925_scan0001.json"
    einputs = [
        {
            "name": "filename",
            "value": "/data/visitor/ls3288/id21/20240221/RAW_DATA/DIV15_10uM_2/DIV15_10uM_2_roi75698_90925/DIV15_10uM_2_roi75698_90925.h5",
            "task_identifier": "PickScan",
        },
        {"name": "scan_number", "value": 1, "task_identifier": "PickScan"},
        {
            "name": "detector_names",
            "value": [
                "fx_nano_det0",
                "fx_nano_det1",
                "fx_nano_det2",
                "fx_nano_det3",
                "fx_nano_det4",
            ],
            "task_identifier": "SumXrfSpectra",
        },
        {
            "name": "output_root_uri",
            "value": "/data/visitor/ls3288/id21/20240221/PROCESSED_DATA/ewoks_results/DIV15_10uM_2_roi75698_90925.h5::/1.1",
            "task_identifier": "SumXrfSpectra",
        },
        {
            "name": "detector_normalization_template",
            "value": "0.1/<instrument/{}/live_time>",
            "task_identifier": "SumXrfSpectra",
        },
        {
            "name": "config",
            "value": "/data/visitor/ls3288/id21/20240221/SCRIPTS/pymca/sample_10200ev_sheldon.cfg",
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "fast_fitting",
            "value": True,
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "quantification",
            "value": True,
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "diagnostics",
            "value": True,
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "energy_name",
            "value": "Edcm",
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "counter_name",
            "value": "niodet",
            "task_identifier": "NormalizeXrfResults",
        },
        {
            "name": "counter_normalization_template",
            "value": "np.nanmean(<instrument/{}/data>)/<instrument/{}/data>",
            "task_identifier": "NormalizeXrfResults",
        },
        {
            "name": "axes_units",
            "value": {"nsy": "mm", "nsz": "mm", "nspy": "um", "nspz": "um"},
            "task_identifier": "RegridXrfResults",
        },
        {
            "name": "ignore_positioners",
            "value": ["nsz1", "nsz2", "nsz3"],
            "task_identifier": "RegridXrfResults",
        },
        {"name": "resolution", "value": None, "task_identifier": "RegridXrfResults"},
    ]

    assert workflow_path == eworkflow_path
    assert convert_destination == econvert_destination
    assert inputs == einputs


@pytest.mark.skipif(
    sys.platform == "win32", reason="Skipping fluo path tests on windows"
)
def test_xrfmap_sum_results():
    workflow_path, inputs, convert_destination = xrfmap_parameters(
        session="/data/visitor/ls3288/id21/20240221",
        sample="DIV15_10uM_2",
        dataset="roi75698_90925",
        scan_number=1,
        config_filenames=[
            "sample_10200ev_sheldon.cfg",
            "sample_10200ev_sheldon.cfg",
            "sample_10200ev_sheldon.cfg",
            "sample_10200ev_sheldon.cfg",
            "sample_10200ev_sheldon.cfg",
        ],
        detector_numbers=[0, 1, 2, 3, 4],
        instrument_name="id21_nano",
        demo=False,
    )

    eworkflow_path = "/data/visitor/ls3288/id21/20240221/SCRIPTS/ewoks_results/workflows/xrfmap_multi_detector.ows"
    econvert_destination = "/data/visitor/ls3288/id21/20240221/PROCESSED_DATA/ewoks_results/DIV15_10uM_2_roi75698_90925_scan0001.json"
    einputs = [
        {
            "name": "filename",
            "value": "/data/visitor/ls3288/id21/20240221/RAW_DATA/DIV15_10uM_2/DIV15_10uM_2_roi75698_90925/DIV15_10uM_2_roi75698_90925.h5",
            "task_identifier": "PickScan",
        },
        {"name": "scan_number", "value": 1, "task_identifier": "PickScan"},
        {
            "name": "output_root_uri",
            "value": "/data/visitor/ls3288/id21/20240221/PROCESSED_DATA/ewoks_results/DIV15_10uM_2_roi75698_90925.h5::/1.1",
            "task_identifier": "FitSingleScanMultiDetector",
        },
        {
            "name": "configs",
            "value": [
                "/data/visitor/ls3288/id21/20240221/SCRIPTS/pymca/sample_10200ev_sheldon.cfg",
                "/data/visitor/ls3288/id21/20240221/SCRIPTS/pymca/sample_10200ev_sheldon.cfg",
                "/data/visitor/ls3288/id21/20240221/SCRIPTS/pymca/sample_10200ev_sheldon.cfg",
                "/data/visitor/ls3288/id21/20240221/SCRIPTS/pymca/sample_10200ev_sheldon.cfg",
                "/data/visitor/ls3288/id21/20240221/SCRIPTS/pymca/sample_10200ev_sheldon.cfg",
            ],
            "task_identifier": "FitSingleScanMultiDetector",
        },
        {
            "name": "fast_fitting",
            "value": True,
            "task_identifier": "FitSingleScanMultiDetector",
        },
        {
            "name": "quantification",
            "value": True,
            "task_identifier": "FitSingleScanMultiDetector",
        },
        {
            "name": "diagnostics",
            "value": True,
            "task_identifier": "FitSingleScanMultiDetector",
        },
        {
            "name": "detector_names",
            "value": [
                "fx_nano_det0",
                "fx_nano_det1",
                "fx_nano_det2",
                "fx_nano_det3",
                "fx_nano_det4",
            ],
            "task_identifier": "FitSingleScanMultiDetector",
        },
        {
            "name": "energy_name",
            "value": "Edcm",
            "task_identifier": "FitSingleScanMultiDetector",
        },
        {
            "name": "detector_normalization_template",
            "value": "0.1/<instrument/{}/live_time>",
            "task_identifier": "SumXrfResults",
        },
        {
            "name": "counter_name",
            "value": "niodet",
            "task_identifier": "NormalizeXrfResults",
        },
        {
            "name": "counter_normalization_template",
            "value": "np.nanmean(<instrument/{}/data>)/<instrument/{}/data>",
            "task_identifier": "NormalizeXrfResults",
        },
        {
            "name": "axes_units",
            "value": {"nsy": "mm", "nsz": "mm", "nspy": "um", "nspz": "um"},
            "task_identifier": "RegridXrfResults",
        },
        {
            "name": "ignore_positioners",
            "value": ["nsz1", "nsz2", "nsz3"],
            "task_identifier": "RegridXrfResults",
        },
        {"name": "resolution", "value": None, "task_identifier": "RegridXrfResults"},
    ]

    assert workflow_path == eworkflow_path
    assert convert_destination == econvert_destination
    assert inputs == einputs


@pytest.mark.skipif(
    sys.platform == "win32", reason="Skipping fluo path tests on windows"
)
def test_mosaic_xrfmap_single_detector():
    workflow_path, inputs, convert_destination = mosaic_xrfmap_parameters(
        session="/data/visitor/blc15972/id21/20250128",
        sample="Cd250",
        datasets=["roi106394_124471", "roi106393_124470"],
        scan_ranges=[(1, 1), (1, 1)],
        config_filenames=["config_9800ev.cfg"],
        detector_numbers=[0],
        instrument_name="id21_nano",
        resolution={"sy": (1, "um"), "sz": (1, "um")},
        demo=False,
    )

    eworkflow_path = "/data/visitor/blc15972/id21/20250128/SCRIPTS/ewoks_results/workflows/mosaic_single_detector.ows"
    econvert_destination = "/data/visitor/blc15972/id21/20250128/PROCESSED_DATA/ewoks_results/Cd250_roi106394_124471_scan0001.json"
    einputs = [
        {
            "name": "filenames",
            "value": [
                "/data/visitor/blc15972/id21/20250128/RAW_DATA/Cd250/Cd250_roi106394_124471/Cd250_roi106394_124471.h5",
                "/data/visitor/blc15972/id21/20250128/RAW_DATA/Cd250/Cd250_roi106393_124470/Cd250_roi106393_124470.h5",
            ],
            "task_identifier": "PickScans",
        },
        {
            "name": "scan_ranges",
            "value": [(1, 1), (1, 1)],
            "task_identifier": "PickScans",
        },
        {"name": "exclude_scans", "value": [(), ()], "task_identifier": "PickScans"},
        {
            "name": "bliss_scan_uri",
            "value": "/data/visitor/blc15972/id21/20250128/PROCESSED_DATA/ewoks_results/Cd250_roi106394_124471_concat.h5::/1.1",
            "task_identifier": "ConcatBliss",
        },
        {
            "name": "output_root_uri",
            "value": "/data/visitor/blc15972/id21/20250128/PROCESSED_DATA/ewoks_results/Cd250_roi106394_124471_results.h5::/1.1",
            "task_identifier": "ConcatBliss",
        },
        {
            "name": "output_root_uri",
            "value": "/data/visitor/blc15972/id21/20250128/PROCESSED_DATA/ewoks_results/Cd250_roi106394_124471_results.h5::/1.1",
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "config",
            "value": "/data/visitor/blc15972/id21/20250128/SCRIPTS/pymca/config_9800ev.cfg",
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "fast_fitting",
            "value": True,
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "quantification",
            "value": True,
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "diagnostics",
            "value": True,
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "detector_name",
            "value": "fx_nano_det0",
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "energy_name",
            "value": "Edcm",
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "counter_name",
            "value": "niodet",
            "task_identifier": "NormalizeXrfResults",
        },
        {
            "name": "counter_normalization_template",
            "value": "np.nanmean(<instrument/{}/data>)/<instrument/{}/data>",
            "task_identifier": "NormalizeXrfResults",
        },
        {
            "name": "detector_normalization_template",
            "value": "0.1/<instrument/{}/live_time>",
            "task_identifier": "NormalizeXrfResults",
        },
        {
            "name": "axes_units",
            "value": {"nsy": "mm", "nsz": "mm", "nspy": "um", "nspz": "um"},
            "task_identifier": "ConcatBliss",
        },
        {
            "name": "virtual_axes",
            "value": {"sy": "<nsy>+<nspy>", "sz": "<nsz>+<nspz>"},
            "task_identifier": "ConcatBliss",
        },
        {
            "name": "positioners",
            "value": ["sz", "sy"],
            "task_identifier": "RegridXrfResults",
        },
        {
            "name": "resolution",
            "value": {"sy": (1, "um"), "sz": (1, "um")},
            "task_identifier": "RegridXrfResults",
        },
    ]

    assert workflow_path == eworkflow_path
    assert convert_destination == econvert_destination
    assert inputs == einputs


@pytest.mark.skipif(
    sys.platform == "win32", reason="Skipping fluo path tests on windows"
)
def test_mosaic_xrfmap_sum_results():
    workflow_path, inputs, convert_destination = mosaic_xrfmap_parameters(
        session="/data/visitor/blc15972/id21/20250128",
        sample="Cd250",
        datasets=["roi106394_124471", "roi106393_124470"],
        scan_ranges=[(1, 1), (1, 1)],
        config_filenames=["config_9800ev.cfg", "config_9800ev.cfg"],
        detector_numbers=[0, 1],
        instrument_name="id21_nano",
        resolution={"sy": (1, "um"), "sz": (1, "um")},
        demo=False,
    )

    eworkflow_path = "/data/visitor/blc15972/id21/20250128/SCRIPTS/ewoks_results/workflows/mosaic_multi_detector.ows"
    econvert_destination = "/data/visitor/blc15972/id21/20250128/PROCESSED_DATA/ewoks_results/Cd250_roi106394_124471_scan0001.json"
    einputs = [
        {
            "name": "filenames",
            "value": [
                "/data/visitor/blc15972/id21/20250128/RAW_DATA/Cd250/Cd250_roi106394_124471/Cd250_roi106394_124471.h5",
                "/data/visitor/blc15972/id21/20250128/RAW_DATA/Cd250/Cd250_roi106393_124470/Cd250_roi106393_124470.h5",
            ],
            "task_identifier": "PickScans",
        },
        {
            "name": "scan_ranges",
            "value": [(1, 1), (1, 1)],
            "task_identifier": "PickScans",
        },
        {"name": "exclude_scans", "value": [(), ()], "task_identifier": "PickScans"},
        {
            "name": "bliss_scan_uri",
            "value": "/data/visitor/blc15972/id21/20250128/PROCESSED_DATA/ewoks_results/Cd250_roi106394_124471_concat.h5::/1.1",
            "task_identifier": "ConcatBliss",
        },
        {
            "name": "output_root_uri",
            "value": "/data/visitor/blc15972/id21/20250128/PROCESSED_DATA/ewoks_results/Cd250_roi106394_124471_results.h5::/1.1",
            "task_identifier": "ConcatBliss",
        },
        {
            "name": "output_root_uri",
            "value": "/data/visitor/blc15972/id21/20250128/PROCESSED_DATA/ewoks_results/Cd250_roi106394_124471_results.h5::/1.1",
            "task_identifier": "FitSingleScanMultiDetector",
        },
        {
            "name": "configs",
            "value": [
                "/data/visitor/blc15972/id21/20250128/SCRIPTS/pymca/config_9800ev.cfg",
                "/data/visitor/blc15972/id21/20250128/SCRIPTS/pymca/config_9800ev.cfg",
            ],
            "task_identifier": "FitSingleScanMultiDetector",
        },
        {
            "name": "fast_fitting",
            "value": True,
            "task_identifier": "FitSingleScanMultiDetector",
        },
        {
            "name": "quantification",
            "value": True,
            "task_identifier": "FitSingleScanMultiDetector",
        },
        {
            "name": "diagnostics",
            "value": True,
            "task_identifier": "FitSingleScanMultiDetector",
        },
        {
            "name": "detector_names",
            "value": ["fx_nano_det0", "fx_nano_det1"],
            "task_identifier": "FitSingleScanMultiDetector",
        },
        {
            "name": "energy_name",
            "value": "Edcm",
            "task_identifier": "FitSingleScanMultiDetector",
        },
        {
            "name": "detector_normalization_template",
            "value": "0.1/<instrument/{}/live_time>",
            "task_identifier": "SumXrfResults",
        },
        {
            "name": "counter_name",
            "value": "niodet",
            "task_identifier": "NormalizeXrfResults",
        },
        {
            "name": "counter_normalization_template",
            "value": "np.nanmean(<instrument/{}/data>)/<instrument/{}/data>",
            "task_identifier": "NormalizeXrfResults",
        },
        {
            "name": "axes_units",
            "value": {"nsy": "mm", "nsz": "mm", "nspy": "um", "nspz": "um"},
            "task_identifier": "ConcatBliss",
        },
        {
            "name": "virtual_axes",
            "value": {"sy": "<nsy>+<nspy>", "sz": "<nsz>+<nspz>"},
            "task_identifier": "ConcatBliss",
        },
        {
            "name": "positioners",
            "value": ["sz", "sy"],
            "task_identifier": "RegridXrfResults",
        },
        {
            "name": "resolution",
            "value": {"sy": (1, "um"), "sz": (1, "um")},
            "task_identifier": "RegridXrfResults",
        },
    ]

    assert workflow_path == eworkflow_path
    assert convert_destination == econvert_destination
    assert inputs == einputs


@pytest.mark.skipif(
    sys.platform == "win32", reason="Skipping fluo path tests on windows"
)
def test_mosaic_xrfmap_sum_spectra():
    workflow_path, inputs, convert_destination = mosaic_xrfmap_parameters(
        session="/data/visitor/blc15972/id21/20250128",
        sample="Cd250",
        datasets=["roi106394_124471", "roi106393_124470"],
        scan_ranges=[(1, 1), (1, 1)],
        config_filenames=["config_9800ev.cfg"],
        detector_numbers=[0, 1],
        instrument_name="id21_nano",
        resolution={"sy": (1, "um"), "sz": (1, "um")},
        demo=False,
    )

    eworkflow_path = "/data/visitor/blc15972/id21/20250128/SCRIPTS/ewoks_results/workflows/mosaic_multi_detector_sumspectra.ows"
    econvert_destination = "/data/visitor/blc15972/id21/20250128/PROCESSED_DATA/ewoks_results/Cd250_roi106394_124471_scan0001.json"
    einputs = [
        {
            "name": "filenames",
            "value": [
                "/data/visitor/blc15972/id21/20250128/RAW_DATA/Cd250/Cd250_roi106394_124471/Cd250_roi106394_124471.h5",
                "/data/visitor/blc15972/id21/20250128/RAW_DATA/Cd250/Cd250_roi106393_124470/Cd250_roi106393_124470.h5",
            ],
            "task_identifier": "PickScans",
        },
        {
            "name": "scan_ranges",
            "value": [(1, 1), (1, 1)],
            "task_identifier": "PickScans",
        },
        {"name": "exclude_scans", "value": [(), ()], "task_identifier": "PickScans"},
        {
            "name": "bliss_scan_uri",
            "value": "/data/visitor/blc15972/id21/20250128/PROCESSED_DATA/ewoks_results/Cd250_roi106394_124471_concat.h5::/1.1",
            "task_identifier": "ConcatBliss",
        },
        {
            "name": "output_root_uri",
            "value": "/data/visitor/blc15972/id21/20250128/PROCESSED_DATA/ewoks_results/Cd250_roi106394_124471_results.h5::/1.1",
            "task_identifier": "ConcatBliss",
        },
        {
            "name": "output_root_uri",
            "value": "/data/visitor/blc15972/id21/20250128/PROCESSED_DATA/ewoks_results/Cd250_roi106394_124471_results.h5::/1.1",
            "task_identifier": "SumXrfSpectra",
        },
        {
            "name": "detector_names",
            "value": ["fx_nano_det0", "fx_nano_det1"],
            "task_identifier": "SumXrfSpectra",
        },
        {
            "name": "detector_normalization_template",
            "value": "0.1/<instrument/{}/live_time>",
            "task_identifier": "SumXrfSpectra",
        },
        {
            "name": "config",
            "value": "/data/visitor/blc15972/id21/20250128/SCRIPTS/pymca/config_9800ev.cfg",
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "fast_fitting",
            "value": True,
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "quantification",
            "value": True,
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "diagnostics",
            "value": True,
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "energy_name",
            "value": "Edcm",
            "task_identifier": "FitSingleScanSingleDetector",
        },
        {
            "name": "counter_name",
            "value": "niodet",
            "task_identifier": "NormalizeXrfResults",
        },
        {
            "name": "counter_normalization_template",
            "value": "np.nanmean(<instrument/{}/data>)/<instrument/{}/data>",
            "task_identifier": "NormalizeXrfResults",
        },
        {
            "name": "axes_units",
            "value": {"nsy": "mm", "nsz": "mm", "nspy": "um", "nspz": "um"},
            "task_identifier": "ConcatBliss",
        },
        {
            "name": "virtual_axes",
            "value": {"sy": "<nsy>+<nspy>", "sz": "<nsz>+<nspz>"},
            "task_identifier": "ConcatBliss",
        },
        {
            "name": "positioners",
            "value": ["sz", "sy"],
            "task_identifier": "RegridXrfResults",
        },
        {
            "name": "resolution",
            "value": {"sy": (1, "um"), "sz": (1, "um")},
            "task_identifier": "RegridXrfResults",
        },
    ]

    assert workflow_path == eworkflow_path
    assert convert_destination == econvert_destination
    assert inputs == einputs


@pytest.mark.skipif(
    sys.platform == "win32", reason="Skipping fluo path tests on windows"
)
def test_fluoxas_single_detector():
    workflow_path, inputs, convert_destination = fluoxas_parameters(
        session="/data/visitor/ma5958/id21/20250207",
        sample="BaCrO4-com-o-aged-thin",
        datasets=["roi107244_125314"],
        scan_ranges=[[2, 3]],
        exclude_scans=[[]],
        config_filenames=["config_nano_6_3.cfg"],
        detector_numbers=[0],
        instrument_name="id21_nano",
        demo=False,
    )

    eworkflow_path = "/data/visitor/ma5958/id21/20250207/SCRIPTS/ewoks_results/workflows/fluoxas_single_detector_align.ows"
    econvert_destination = "/data/visitor/ma5958/id21/20250207/PROCESSED_DATA/ewoks_results/BaCrO4-com-o-aged-thin_roi107244_125314_scan0002.json"
    einputs = [
        {
            "name": "filenames",
            "value": [
                "/data/visitor/ma5958/id21/20250207/RAW_DATA/BaCrO4-com-o-aged-thin/BaCrO4-com-o-aged-thin_roi107244_125314/BaCrO4-com-o-aged-thin_roi107244_125314.h5"
            ],
            "task_identifier": "PickScans",
        },
        {"name": "scan_ranges", "value": [(2, 3)], "task_identifier": "PickScans"},
        {"name": "exclude_scans", "value": [[]], "task_identifier": "PickScans"},
        {
            "name": "output_root_uri",
            "value": "/data/visitor/ma5958/id21/20250207/PROCESSED_DATA/ewoks_results/BaCrO4-com-o-aged-thin_roi107244_125314.h5::/2.1",
            "task_identifier": "FitStackSingleDetector",
        },
        {
            "name": "config",
            "value": "/data/visitor/ma5958/id21/20250207/SCRIPTS/pymca/config_nano_6_3.cfg",
            "task_identifier": "FitStackSingleDetector",
        },
        {
            "name": "fast_fitting",
            "value": True,
            "task_identifier": "FitStackSingleDetector",
        },
        {
            "name": "quantification",
            "value": True,
            "task_identifier": "FitStackSingleDetector",
        },
        {
            "name": "diagnostics",
            "value": True,
            "task_identifier": "FitStackSingleDetector",
        },
        {
            "name": "detector_name",
            "value": "fx_nano_det0",
            "task_identifier": "FitStackSingleDetector",
        },
        {
            "name": "energy_name",
            "value": "Edcm",
            "task_identifier": "FitStackSingleDetector",
        },
        {
            "name": "counter_name",
            "value": "niodet",
            "task_identifier": "NormalizeXrfResultsStack",
        },
        {
            "name": "counter_normalization_template",
            "value": "np.nanmean(<instrument/{}/data>)/<instrument/{}/data>",
            "task_identifier": "NormalizeXrfResultsStack",
        },
        {
            "name": "detector_normalization_template",
            "value": "0.1/<instrument/{}/live_time>",
            "task_identifier": "NormalizeXrfResultsStack",
        },
        {
            "name": "axes_units",
            "value": {"nsy": "mm", "nsz": "mm", "nspy": "um", "nspz": "um"},
            "task_identifier": "RegridXrfResultsStack",
        },
        {
            "name": "ignore_positioners",
            "value": ["nsz1", "nsz2", "nsz3"],
            "task_identifier": "RegridXrfResultsStack",
        },
        {
            "name": "resolution",
            "value": None,
            "task_identifier": "RegridXrfResultsStack",
        },
        {
            "name": "reference_stack",
            "value": None,
            "task_identifier": "Reg2DPreEvaluation",
        },
        {"name": "skip", "value": True, "task_identifier": "Reg2DPreEvaluation"},
        {"name": "skip", "value": False, "task_identifier": "Reg2DPostEvaluation"},
        {"name": "crop", "value": True, "task_identifier": "Reg2DTransform"},
    ]

    assert workflow_path == eworkflow_path
    assert convert_destination == econvert_destination
    assert inputs == einputs


@pytest.mark.skipif(
    sys.platform == "win32", reason="Skipping fluo path tests on windows"
)
def test_fluoxas_sum_results():
    workflow_path, inputs, convert_destination = fluoxas_parameters(
        session="/data/visitor/ma5958/id21/20250207",
        sample="BaCrO4-com-o-aged-thin",
        datasets=["roi107244_125314"],
        scan_ranges=[[2, 3]],
        exclude_scans=[[]],
        config_filenames=["config_nano_6_3.cfg", "config_nano_6_3.cfg"],
        detector_numbers=[0, 1],
        instrument_name="id21_nano",
        demo=False,
    )

    eworkflow_path = "/data/visitor/ma5958/id21/20250207/SCRIPTS/ewoks_results/workflows/fluoxas_multi_detector_align.ows"
    econvert_destination = "/data/visitor/ma5958/id21/20250207/PROCESSED_DATA/ewoks_results/BaCrO4-com-o-aged-thin_roi107244_125314_scan0002.json"
    einputs = [
        {
            "name": "filenames",
            "value": [
                "/data/visitor/ma5958/id21/20250207/RAW_DATA/BaCrO4-com-o-aged-thin/BaCrO4-com-o-aged-thin_roi107244_125314/BaCrO4-com-o-aged-thin_roi107244_125314.h5"
            ],
            "task_identifier": "PickScans",
        },
        {"name": "scan_ranges", "value": [(2, 3)], "task_identifier": "PickScans"},
        {"name": "exclude_scans", "value": [[]], "task_identifier": "PickScans"},
        {
            "name": "output_root_uri",
            "value": "/data/visitor/ma5958/id21/20250207/PROCESSED_DATA/ewoks_results/BaCrO4-com-o-aged-thin_roi107244_125314.h5::/2.1",
            "task_identifier": "FitStackMultiDetector",
        },
        {
            "name": "configs",
            "value": [
                "/data/visitor/ma5958/id21/20250207/SCRIPTS/pymca/config_nano_6_3.cfg",
                "/data/visitor/ma5958/id21/20250207/SCRIPTS/pymca/config_nano_6_3.cfg",
            ],
            "task_identifier": "FitStackMultiDetector",
        },
        {
            "name": "fast_fitting",
            "value": True,
            "task_identifier": "FitStackMultiDetector",
        },
        {
            "name": "quantification",
            "value": True,
            "task_identifier": "FitStackMultiDetector",
        },
        {
            "name": "diagnostics",
            "value": True,
            "task_identifier": "FitStackMultiDetector",
        },
        {
            "name": "detector_names",
            "value": ["fx_nano_det0", "fx_nano_det1"],
            "task_identifier": "FitStackMultiDetector",
        },
        {
            "name": "energy_name",
            "value": "Edcm",
            "task_identifier": "FitStackMultiDetector",
        },
        {
            "name": "detector_normalization_template",
            "value": "0.1/<instrument/{}/live_time>",
            "task_identifier": "SumXrfResultsStack",
        },
        {
            "name": "counter_name",
            "value": "niodet",
            "task_identifier": "NormalizeXrfResultsStack",
        },
        {
            "name": "counter_normalization_template",
            "value": "np.nanmean(<instrument/{}/data>)/<instrument/{}/data>",
            "task_identifier": "NormalizeXrfResultsStack",
        },
        {
            "name": "axes_units",
            "value": {"nsy": "mm", "nsz": "mm", "nspy": "um", "nspz": "um"},
            "task_identifier": "RegridXrfResultsStack",
        },
        {
            "name": "ignore_positioners",
            "value": ["nsz1", "nsz2", "nsz3"],
            "task_identifier": "RegridXrfResultsStack",
        },
        {
            "name": "resolution",
            "value": None,
            "task_identifier": "RegridXrfResultsStack",
        },
        {
            "name": "reference_stack",
            "value": None,
            "task_identifier": "Reg2DPreEvaluation",
        },
        {"name": "skip", "value": True, "task_identifier": "Reg2DPreEvaluation"},
        {"name": "skip", "value": False, "task_identifier": "Reg2DPostEvaluation"},
        {"name": "crop", "value": True, "task_identifier": "Reg2DTransform"},
    ]

    assert workflow_path == eworkflow_path
    assert convert_destination == econvert_destination
    assert inputs == einputs



from typing import Any
from typing import Dict
from typing import Optional

from ..id09.txs_processor import TxsProcessor


class DemoTxsProcessor(TxsProcessor):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        super().__init__(config=config, defaults=defaults)

        self.detector = "difflab6"
        self.pixel = (10e-3, 10e-3)
        self.energy = 10


txs_processor = DemoTxsProcessor()



import numpy
import pytest

from ..id06.utils import get_positions
from ..id06.xrpd_processor import Id06XrpdProcessor

PI = numpy.pi


def test_get_positions():
    deg_step = 1
    rad_step = numpy.deg2rad(deg_step)

    pyfai_pos = get_positions(start=0, step=1, npts=360)
    theoretical_pos = numpy.concatenate(
        (
            numpy.arange(-PI + rad_step / 2, 0, rad_step),
            numpy.arange(0, PI, rad_step),
        ),
        axis=0,
    )

    assert numpy.all(pyfai_pos - theoretical_pos < 1e-13)


class MockScan:
    def __init__(self, scan_info):
        self.scan_info = scan_info


@pytest.mark.parametrize(
    "scan_info",
    [
        ({"type": "fscan", "save": True}),
        ({"type": "oscan", "save": True}),
    ],
)
def test_get_integration_options_parametrized(mock_bliss, scan_info):
    mock_scan = MockScan(scan_info=scan_info)
    id06_xrpd_processor = Id06XrpdProcessor()
    id06_xrpd_processor.integration_options = {"param": 22}

    result = id06_xrpd_processor.get_integration_options(mock_scan, "lima_name")

    assert result["param"] == 22



import functools
import inspect
import pathlib

from ._display import print_message_on_exit


def demo_assert(message: str):
    if not isinstance(message, str):
        raise TypeError(f"{message} is not a string")

    def decorator(test_method):
        @functools.wraps(test_method)
        def wrapper(*args, **kwargs):
            bound_args = inspect.signature(test_method).bind(*args, **kwargs)
            bound_args.apply_defaults()
            arguments = {
                k: repr(str(v)) if isinstance(v, (pathlib.Path, str)) else v
                for k, v in bound_args.arguments.items()
            }
            msg = message.format(**arguments)

            with print_message_on_exit(msg, "assert"):
                return test_method(*args, **kwargs)

        return wrapper

    return decorator



"""For the BLISS demo session.

* `blissoda.demo.processors`: Processor classes adapted for the Bliss demo session.
* `blissoda.demo.testing`: pytest-like helper functions to create integration tests.
  * See `blissoda.demo.scripts.template` on how they are intended.
* `blissoda.demo.scripts`: Bliss users scripts
  * To be loaded with `user_script_load` after `user_script_homedir_oda`.
  * Use `blissoda.demo.testing` to create integration tests.
"""

import os

EWOKS_EVENTS_DIR = os.path.join(
    os.path.abspath(os.environ.get("DEMO_TMP_ROOT", ".")), "ewoks_events"
)

EWOKS_RESULTS_DIR = os.path.join(
    os.path.abspath(os.environ.get("DEMO_TMP_ROOT", ".")), "ewoks_results"
)



from blissoda.bliss_globals import setup_globals
from blissoda.demo.id32 import id32_processor


def id32_demo(expo=0.2, npoints=10):
    print("The workflow is expected to fail so this only tests the triggering.")
    id32_processor.enable()
    try:
        setup_globals.loopscan(npoints, expo, setup_globals.difflab6)
    finally:
        id32_processor.disable()



from ewoksjob.client import Future

from ..bliss_globals import current_session
from ..import_utils import unavailable_module

try:
    import gevent
except ImportError as ex:
    gevent = unavailable_module(ex)


def send_figures_to_elogbook(future: Future, retry_timeout: int, retry_period: int):
    result = future.result(timeout=retry_timeout, interval=retry_period)

    saved_figs = result.get("saved_figs", None)
    if saved_figs is None:
        return

    icat_client = current_session.scan_saving.icat_client
    for saved_fig in saved_figs:
        icat_client.send_binary_file(saved_fig)


class CdiUploader:
    def __init__(self) -> None:
        super().__init__()
        self._tasks = []

    def upload_workflow_result(
        self, future: Future, retry_timeout: int, retry_period: int
    ):
        task = gevent.spawn(
            send_figures_to_elogbook, future, retry_timeout, retry_period
        )
        self._tasks.append(task)
        self.purge_tasks()

    def purge_tasks(self) -> int:
        """Remove references to tasks that have finished."""
        self._tasks = [t for t in self._tasks if t]
        return len(self._tasks)

    def kill_tasks(self) -> int:
        """Kill all tasks."""
        gevent.killall(self._tasks)
        return self.purge_tasks()



"""Automatic pyfai integration for every scan with saving and plotting"""

import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ewoksutils.task_utils import task_inputs

from ..bliss_globals import setup_globals
from ..import_utils import unavailable_class
from ..persistent.parameters import ParameterInfo
from ..processor import BlissScanType
from ..utils import validators
from ..xrpd.processor import XrpdProcessor

try:
    from bliss.common.axis import Axis
except ImportError as ex:
    Axis = unavailable_class(ex)


def is_motor_name(value) -> str:
    if isinstance(value, Axis):
        return str(value.name)
    return str(value)


class Id31XrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("pyfai_config", category="PyFai", validator=validators.is_file),
        ParameterInfo("integration_options", category="PyFai"),
        ParameterInfo("flat_enabled", category="Flat-field", validator=bool),
        ParameterInfo("newflat", category="Flat-field", validator=validators.is_file),
        ParameterInfo("oldflat", category="Flat-field", validator=validators.is_file),
        ParameterInfo(
            "ascii_options",
            category="Data saving",
        ),
        ParameterInfo(
            "tomo_enabled",
            category="Data saving/Tomo",
            validator=bool,
        ),
        ParameterInfo(
            "tomo_rot_name",
            category="Data saving/Tomo",
            validator=is_motor_name,
        ),
        ParameterInfo(
            "tomo_y_name",
            category="Data saving/Tomo",
            validator=is_motor_name,
        ),
        ParameterInfo(
            "tomo_scans",
            category="Data saving/Tomo",
        ),
    ],
):

    # Override XrpdProcessor default workflows
    DEFAULT_WORKFLOW = "integrate_with_saving_with_flat.json"
    DEFAULT_WORKFLOW_NO_SAVE = "integrate_without_saving_with_flat.json"

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("queue", "online")
        defaults.setdefault(
            "integration_options",
            {
                "method": "no_csr_ocl_gpu",
                "nbpt_rad": 4096,
                "unit": "q_nm^-1",
            },
        )
        defaults.setdefault(
            "ascii_options",
            {
                "enabled": False,
                "scans": [
                    "ascan",
                    "dscan",
                    "a2scan",
                    "d2scan",
                    "loopscan",
                    "timescan",
                    "sct",
                ],
                "max_npoints": 1000,
            },
        )
        defaults.setdefault("flat_enabled", True)
        defaults.setdefault("newflat", "/data/id31/inhouse/P3/flats.mat")
        defaults.setdefault("oldflat", "/data/id31/inhouse/P3/flats_old.mat")
        defaults.setdefault("tomo_enabled", False)
        defaults.setdefault("tomo_rot_name", "srot")
        defaults.setdefault("tomo_y_name", "saby")
        defaults.setdefault("tomo_scans", ["fscan2d"])

        super().__init__(config=config, defaults=defaults)

        self.workflow_with_saving = self.DEFAULT_WORKFLOW
        self.workflow_without_saving = self.DEFAULT_WORKFLOW_NO_SAVE

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        return self.pyfai_config

    def get_integration_options(
        self, scan: BlissScanType, lima_name: str
    ) -> Optional[dict]:
        integration_options = self.integration_options
        if integration_options:
            return integration_options.to_dict()
        return None

    def get_inputs(self, scan, lima_name: str) -> List[dict]:
        inputs = super().get_inputs(scan, lima_name)
        inputs += task_inputs(
            task_identifier="FlatFieldFromEnergy",
            inputs={
                "newflat": self.newflat,
                "oldflat": self.oldflat,
                "energy": setup_globals.energy.position,
                "enabled": self.flat_enabled and lima_name == "p3",
            },
        )

        ascii_basename_template = self._export_filename_prefix(scan) + "_%04d.xye"
        if scan.scan_info.get("npoints", float("inf")) == 1:
            output_filename_template = os.path.join(
                self._export_folder(scan), ascii_basename_template
            )
            output_archive_filename = ""
        else:  # Store ASCII files in a zip file if there is more than one point
            output_filename_template = ascii_basename_template
            output_archive_filename = os.path.join(
                self._export_folder(scan),
                self._export_filename_prefix(scan) + ".zip",
            )
        inputs += task_inputs(
            task_identifier="SaveNexusPatternsAsAscii",
            inputs={
                "enabled": self._is_save_ascii_enabled(scan),
                "output_filename_template": output_filename_template,
                "output_archive_filename": output_archive_filename,
            },
        )

        output_tomo_filename = os.path.join(
            self._export_folder(scan),
            self._export_filename_prefix(scan) + "_tomo.h5",
        )
        scan_nb = scan.scan_info.get("scan_nb")
        inputs += task_inputs(
            task_identifier="SaveNexusPatternsAsId31TomoHdf5",
            inputs={
                "enabled": self._is_save_tomo_enabled(scan),
                "scan_entry_url": f"{self.get_filename(scan)}::/{scan_nb}.1",
                "rot_name": self.tomo_rot_name,
                "y_name": self.tomo_y_name,
                "output_filename": output_tomo_filename,
            },
        )
        return inputs

    def _is_save_tomo_enabled(self, scan) -> bool:
        """Returns whether results need to be exported as tomo for this scan"""
        if not self.tomo_enabled:
            return False

        scan_type = scan.scan_info.get("type", "")
        return scan_type in self.tomo_scans

    def _is_save_ascii_enabled(self, scan) -> bool:
        """Returns whether results need to be exported as ASCII for this scan"""
        if not self.ascii_options:
            return False

        ascii_options = self.ascii_options.to_dict()
        if not ascii_options.get("enabled", False):
            return False

        scan_type = scan.scan_info.get("type", "")
        # sct's scan_type is ct with saving enabled
        if scan_type == "ct" and scan.scan_info.get("save"):
            scan_type = "sct"

        return scan_type in ascii_options.get("scans", []) and (
            0
            < scan.scan_info.get("npoints", float("inf"))
            <= ascii_options.get("max_npoints", 0)
        )

    def _export_filename_prefix(self, scan) -> str:
        """Returns the prefix to use for exported filenames"""
        basename = os.path.basename(self.get_filename(scan))
        stem = os.path.splitext(basename)[0]
        scan_nb = scan.scan_info.get("scan_nb")
        return f"{stem}_{scan_nb:04d}"

    def _export_folder(self, scan) -> str:
        """Returns the folder where to store exported files"""
        return os.path.join(
            self.scan_processed_directory(scan),
            "export",
        )

    def get_workflow(self, scan) -> Optional[str]:
        return self._get_workflow(scan)

    def _get_default_workflow(self, scan) -> Optional[str]:
        # Make sure default workflows are never used
        return self._get_workflow(scan)

    def get_submit_arguments(self, scan, lima_name) -> dict:
        kwargs = super().get_submit_arguments(scan, lima_name)
        kwargs["outputs"] = [
            {"all": False},
            {"name": "nxdata_url", "task_identifier": "IntegrateBlissScan"},
        ]
        kwargs["load_options"] = {"root_module": "ewoksid31.workflows"}
        return kwargs

        # TODO: Redis events don't show up
        handler = {
            "class": "ewoksjob.events.handlers.RedisEwoksEventHandler",
            "arguments": [
                {
                    "name": "url",
                    "value": "redis://bibhelm:25001/4",
                },
                {"name": "ttl", "value": 86400},
            ],
        }
        kwargs["execinfo"] = {"handlers": [handler]}
        return kwargs

    def enabled_flatfield(self, enable: bool) -> None:
        # Kept for backward compatibility
        self.flat_enabled = enable

    def ensure_workflow_accessible(self, scan) -> None:
        pass



import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ewoksjob.client import submit

from ..persistent.parameters import ParameterInfo
from ..processor import BaseProcessor
from ..processor import BlissScanType

_DEFAULT_SCAN_PARS: Dict[str, Any] = {}


class TiffFilesProcessor(
    BaseProcessor,
    parameters=[
        ParameterInfo("queue", category="workflows", deprecated_names=["worker"]),
        ParameterInfo("workflow", category="workflows"),
        ParameterInfo(
            "lima_name",
            category="TiffFiles",
            doc="Name of the detector used for the conversion",
        ),
        ParameterInfo(
            "scan_parameters",
            category="TiffFiles",
            doc="Derived from scan_info",
        ),
    ],
):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        # default trigger at end of scan
        defaults.setdefault("trigger_at", "END")
        defaults.setdefault("workflow", "")
        defaults.setdefault("lima_name", "eiger")
        # initialize empty scan parameters
        defaults.setdefault("scan_parameters", _DEFAULT_SCAN_PARS)

        super().__init__(config=config, defaults=defaults)

    def update_scan_parameters(self, scan: BlissScanType = None) -> Dict[str, Any]:
        scan_parameters: Dict[str, Any] = {}
        if scan is not None:
            # HDF5 input file from LIMA
            scan_number = scan.scan_number
            image_path = scan.scan_saving.images_path.format(
                scan_number=scan_number,
                img_acq_device=self.lima_name if hasattr(self, "lima_name") else "",
            )
            # typically the first frame file
            scan_parameters["images"] = [f"{image_path}0000.h5"]
            # output base name and folder
            processed_dir = os.path.dirname(
                scan_parameters["images"][0].replace("RAW_DATA", "PROCESSED_DATA")
            )
            # folder where TIFFs will be created
            output_folder = os.path.join((processed_dir), "xdi")
            scan_parameters["output"] = output_folder
        self.scan_parameters.update(scan_parameters)
        return scan_parameters

    def get_inputs(self, scan: BlissScanType) -> List[Dict[str, Any]]:
        # ensure scan_parameters are up to date
        params = self.update_scan_parameters(scan)
        inputs: List[Dict[str, Any]] = [
            {
                "task_identifier": "TiffFiles",
                "name": "images",
                "value": params["images"],
            },
            {
                "task_identifier": "TiffFiles",
                "name": "output",
                "value": params["output"],
            },
        ]
        if self.lima_name:
            inputs.append(
                {
                    "task_identifier": "TiffFiles",
                    "name": "detector_name",
                    "value": self.lima_name,
                }
            )
        return inputs

    def workflow_destination(self, scan) -> str:
        scan_number = scan.scan_info.get("scan_nb", 0)
        dataset_name = os.path.basename(os.path.dirname(scan.scan_saving.filename))
        return os.path.join(
            self.scan_parameters["output"], f"{dataset_name}_{scan_number}.json"
        )

    def get_submit_arguments(self, scan: BlissScanType) -> Dict[str, Any]:
        return {"inputs": self.get_inputs(scan), "outputs": [{"all": "False"}]}

    def run_conversion(self, scan: BlissScanType) -> None:
        kw = self.get_submit_arguments(scan)
        # destination for workflow metadata
        kw["convert_destination"] = self.workflow_destination(scan)
        submit(args=(self.workflow,), kwargs=kw, queue=self.queue)

    def _trigger_workflow_on_new_scan(self, scan: BlissScanType) -> None:
        # trigger only for fscan-type scans
        if "dmesh" in scan.scan_info.get("type", ""):
            self.run_conversion(scan)



from typing import Any
from typing import Callable
from typing import Type


def unavailable_function(ex: Exception) -> Callable[..., Any]:
    """Return a function that raises the given exception when called."""

    def _raiser(*args: Any, **kwargs: Any) -> Any:
        raise ex

    _raiser.__unavailable__ = True
    return _raiser


class UnavailableObject:
    """Object placeholder that raises an exception on any attribute access."""

    __unavailable__ = True

    def __init__(self, ex: Exception) -> None:
        self.__ex = ex

    def __getattr__(self, name: str) -> Any:
        if name in ("__func__", "_is_coroutine_marker", "_is_coroutine"):
            # Allow monkey patching for tests
            return super().__getattr__(name)
        raise self.__ex


def unavailable_class(ex: Exception) -> Type[Any]:
    """Return a class that raises the given exception on instantiation."""

    class _Unavailable:
        __unavailable__ = True

        def __init__(self, *args: Any, **kwargs: Any) -> None:
            raise ex

    return _Unavailable


def unavailable_type(ex: Exception) -> Type[Any]:
    """Return a type placeholder that raises the given exception on instantiation."""
    return unavailable_class(ex)


def unavailable_module(ex: Exception) -> Any:
    """Return an object acting like an unavailable module, raising on attribute access."""
    return UnavailableObject(ex)


def is_available(obj: Any) -> bool:
    """Check if an object is marked as unavailable."""
    return not getattr(obj, "__unavailable__", False)



"""User API for HDF5 conversion on the Bliss repl"""

import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ewoksjob.client import submit

from ..bliss_globals import current_session
from ..persistent.parameters import ParameterInfo
from ..processor import BaseProcessor
from ..processor import BlissScanType
from ..utils import directories


class id16bXeol(
    BaseProcessor,
    parameters=[
        ParameterInfo("workflow", category="workflows"),
        ParameterInfo("retry_timeout", category="data access"),
        ParameterInfo("queue", category="workflows", deprecated_names=["worker"]),
        ParameterInfo("calibration_x"),
        ParameterInfo("calibration_y"),
        ParameterInfo("threshold"),
        ParameterInfo("configuration_path"),
        ParameterInfo("h5name"),
        ParameterInfo("save_path"),
        ParameterInfo("normalization_counter"),
        ParameterInfo("main_counter"),
        ParameterInfo("pixels"),
    ],
):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
    ) -> None:
        if defaults is None:
            defaults = {}
        defaults.setdefault("trigger_at", "END")
        defaults.setdefault("queue", "celery")
        defaults.setdefault("calibration_x", [0, 1])
        defaults.setdefault("calibration_y", [0, 1])
        defaults.setdefault("threshold", 0)
        defaults.setdefault("pixels", None)
        current_dir = os.path.dirname(os.path.abspath(__file__))
        workflow_path = os.path.join(
            current_dir, "..", "resources/id16b/id16b_xeol.json"
        )
        defaults.setdefault("workflow", workflow_path)
        super().__init__(config=config, defaults=defaults)

    def on_new_scan_metadata(self, scan: BlissScanType) -> None:
        if not self.scan_requires_processing(scan):
            return
        kwargs = self.get_submit_arguments(scan)
        _ = submit(args=(self.workflow,), kwargs=kwargs, queue=self.queue)

    def _trigger_workflow_on_new_scan(self, scan: BlissScanType) -> None:
        self.on_new_scan_metadata(scan)

    def scan_requires_processing(self, scan: BlissScanType) -> bool:
        if self.find_in_dict(scan.scan_info, self.main_counter):
            return True
        return False

    def find_in_dict(self, current_dict, key):
        try:
            if key == current_dict["display_name"]:
                return True
        except (KeyError, AttributeError):
            for value in current_dict.values():
                if isinstance(value, dict) and self.find_in_dict(value, key):
                    return True
        return False

    def get_submit_arguments(self, scan: BlissScanType) -> dict:
        return {
            "inputs": self.get_inputs(scan),
            "outputs": [{"all": False}],
        }

    def get_filename(self, scan: BlissScanType) -> str:
        filename = scan.scan_info.get("filename")
        if filename:
            return filename
        return current_session.scan_saving.filename

    def workflow_destination(self, scan: BlissScanType) -> str:
        filename = self.get_filename(scan)
        root = directories.get_processed_dir(filename)
        basename = "xeol.h5"
        return os.path.join(root, basename)

    def get_inputs(self, scan: BlissScanType) -> List[dict]:
        # do not read data from file "self.get_filename(scan)" there are issues
        # see id16b ewoks tasks for details
        filename = self.get_filename(scan)
        scan_id = scan._scan_data.key  # self.redis_key
        if self.save_path is None:
            output_filename = self.workflow_destination(scan)
        else:
            output_filename = self.save_path
        scan_nb = scan.scan_info.get("scan_nb")
        if output_filename and filename:
            inputs = [
                {
                    "name": "save_path",
                    "value": output_filename,
                    "task_identifier": "XeolStackFit",
                },
                {
                    "name": "stack_path",
                    "value": filename,
                    "task_identifier": "ReadStack",
                },
                {
                    "name": "stack_path",
                    "value": filename,
                    "task_identifier": "SaveXeolH5",
                },
                {"name": "key", "value": scan_id, "task_identifier": "ReadStack"},
                {"name": "key", "value": scan_id, "task_identifier": "SaveXeolH5"},
                {
                    "name": "scan_number",
                    "value": scan_nb,
                    "task_identifier": "ReadStack",
                },
                {
                    "name": "scan_number",
                    "value": scan_nb,
                    "task_identifier": "SaveXeolH5",
                },
            ]
        else:
            raise Exception("could not read save path and/or read path")

        if self.configuration_path:
            inputs.append(
                {
                    "name": "config_path",
                    "value": self.configuration_path,
                    "task_identifier": "ReadCorrectConfig",
                },
            )
        else:
            raise Exception(
                "Configuration path is required. Please set it before start"
            )

        if self.calibration_x is None:
            self.calibration_x = scan.scan_info.get("instrument", dict()).get(
                "calibration"
            )
        if self.calibration_x:
            inputs.append(
                {
                    "name": "calibration_x",
                    "value": self.calibration_x,
                    "task_identifier": "ReadStack",
                },
            )
            inputs.append(
                {
                    "name": "calibration_x",
                    "value": self.calibration_x,
                    "task_identifier": "SaveXeolH5",
                },
            )

        if self.calibration_y:
            inputs.append(
                {
                    "name": "calibration_y",
                    "value": self.calibration_y,
                    "task_identifier": "ReadStack",
                },
            )

        if self.h5name:
            inputs.append(
                {
                    "name": "h5name",
                    "value": self.h5name,
                    "task_identifier": "SaveXeolH5",
                },
            )

        if self.normalization_counter:
            inputs.append(
                {
                    "name": "norm_counter",
                    "value": self.normalization_counter,
                    "task_identifier": "ReadStack",
                },
            )

        if self.threshold:
            inputs.append(
                {
                    "name": "threshold",
                    "value": self.threshold,
                    "task_identifier": "ReadStack",
                },
            )

        if self.main_counter:
            inputs.append(
                {
                    "name": "counter",
                    "value": self.main_counter,
                    "task_identifier": "ReadStack",
                },
            )

        if self.pixels:
            inputs.append(
                {
                    "name": "pixels",
                    "value": self.pixels,
                    "task_identifier": "ReadStack",
                },
            )

        return inputs

    @property
    def calibration_x(self):
        return self._get_parameter("calibration_x")

    @property
    def calibration_y(self):
        return self._get_parameter("calibration_y")

    @property
    def threshold(self):
        return self._get_parameter("threshold")

    @property
    def configuration_path(self):
        return self._get_parameter("configuration_path")

    @property
    def h5name(self):
        return self._get_parameter("h5name")

    @property
    def normalization_counter(self):
        return self._get_parameter("normalization_counter")

    @property
    def main_counter(self):
        return self._get_parameter("main_counter")

    @property
    def save_path(self):
        return self._get_parameter("save_path")

    @property
    def pixels(self):
        return self._get_parameter("pixels")

    @calibration_x.setter
    def calibration_x(self, *args):
        if len(args) == 1:
            if isinstance(args[0], list) and len(args[0]) >= 2:
                self._set_parameter("calibration_x", args[0])
            else:
                raise Exception(
                    "Wrong format for calibration. Please insert a list or a sequence longer than 2"
                )
        elif len(args) >= 2:
            self._set_parameter("calibration_x", list(args))
        else:
            raise Exception(
                "Wrong format for calibration. Please insert a list or a sequence longer than 2"
            )

    @calibration_y.setter
    def calibration_y(self, *args):
        if len(args) == 1:
            if isinstance(args[0], list) and len(args[0]) >= 2:
                self._set_parameter("calibration_y", args[0])
            else:
                raise Exception(
                    "Wrong format for calibration. Please insert a list or a sequence longer than 2"
                )
        elif len(args) >= 2:
            self._set_parameter("calibration_y", list(args))
        else:
            raise Exception(
                "Wrong format for calibration. Please insert a list or a sequence longer than 2"
            )

    @threshold.setter
    def threshold(self, lower_limit):
        if lower_limit < 0:
            raise Exception("Lower limit cannot be negative")
        else:
            self._set_parameter("threshold", lower_limit)

    @configuration_path.setter
    def configuration_path(self, path):
        if os.path.exists(path):
            self._set_parameter("configuration_path", path)
        else:
            raise Exception("This path does not exist")

    @h5name.setter
    def h5name(self, filename):
        self._set_parameter("h5name", filename)

    @normalization_counter.setter
    def normalization_counter(self, counter):
        self._set_parameter("normalization_counter", counter)

    @main_counter.setter
    def main_counter(self, counter):
        self._set_parameter("main_counter", counter)

    @save_path.setter
    def save_path(self, path):
        if os.path.isdir(path):
            self._set_parameter("save_path", path)
        else:
            raise Exception("This folder does not exist. PLease create it manually.")

    @pixels.setter
    def pixels(self, pix):
        if isinstance(pix, list) and len(pix) == 2:
            self._set_parameter("pixels", pix)
        else:
            raise Exception("Inappropriate format")



import json
import logging
import os
import pathlib
import time
from typing import Any
from typing import Dict
from typing import Optional

from ewoksjob.client import submit

from ..bliss_globals import current_session
from ..flint.access import WithFlintAccess
from ..import_utils import unavailable_function
from ..import_utils import unavailable_module
from ..persistent.parameters import ParameterInfo
from ..persistent.parameters import autocomplete_property
from ..processor import BaseProcessor
from ..processor import BlissScanType
from ..resources import resource_filename
from ..utils.directories import get_dataset_processed_dir
from .plots import SpecGenPlot

try:
    import gevent
except ImportError as ex:
    gevent = unavailable_module(ex)

try:
    from bliss.shell.getval import getval_float
except ImportError as ex:
    getval_float = unavailable_function(ex)


try:
    from bliss.shell.getval import getval_int
except ImportError as ex:
    getval_int = unavailable_function(ex)


try:
    from bliss.shell.getval import getval_yes_no
except ImportError as ex:
    getval_yes_no = unavailable_function(ex)

logger = logging.getLogger(__name__)


class Id32SpecGenProcessor(
    BaseProcessor,
    parameters=[
        ParameterInfo("parameters", category="workflows"),
        ParameterInfo("max_scans_flint", category="plotting"),
        ParameterInfo("update_period", category="workflows"),
        ParameterInfo("update_period_flint", category="plotting"),
    ],
    deprecated_class_attributes={"WORKER": "QUEUE"},
):
    QUEUE = "lid32xmcd2"
    WORKFLOW_FILENAME = "convert_image_to_spectrum.json"

    def __init__(
        self,
        detectors=[],
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault(
            "parameters",
            {
                detector: {
                    "slope": -0.0688,
                    "smile": float(0),
                    "energy calibration (meV/px)": 21.5,
                    "points per pixel": 2.7,
                    "low threshold": 0.12,
                    "high threshold": 1,
                    "mask size": 5,
                    "SPC": True,
                    "SPC grid size": 3,
                    "SPC low threshold": 0.2,
                    "SPC high threshold": 1.0,
                    "SPC single event threshold": 0.4,
                    "SPC double event threshold": 1.5,
                }
                for detector in detectors
            },
        )
        defaults.setdefault("max_scans_flint", 5)
        defaults.setdefault("update_period_flint", 1)
        defaults.setdefault("update_period", 60)

        self.detectors = detectors

        super().__init__(config=config, defaults=defaults)

        self._plotters = {
            detector: Plotter(
                max_scans=self.max_scans_flint,
                update_period=self.update_period_flint,
                detector=detector,
            )
            for detector in self.detectors
        }

    @autocomplete_property
    def max_scans_flint(self) -> Optional[int]:
        return self._get_parameter("max_scans_flint")

    @max_scans_flint.setter
    def max_scans_flint(self, value: int):
        for plotter in self._plotters:
            self._plotters[plotter].max_scans = value
            self._set_parameter("max_scans_flint", self._plotters[plotter].max_scans)

    @autocomplete_property
    def update_period_flint(self) -> Optional[float]:
        return self._get_parameter("update_period_flint")

    @update_period_flint.setter
    def update_period_flint(self, value: float):
        for plotter in self._plotters:
            self._plotters[plotter].update_period = value
            self._set_parameter(
                "update_period_flint", self._plotters[plotter].update_period
            )

    def setup(self):
        req = "  ".join(
            ["(%d) %s" % (i + 1, det) for i, det in enumerate(self.detectors)]
        )
        if not req:
            print("No detector defined")
            return

        ret = getval_int(
            "Detector?  " + req, minimum=1, maximum=len(self.detectors), default=1
        )
        det = self.detectors[ret - 1]
        for par in self.parameters[det].keys():
            old_par = self.parameters[det][par]
            if isinstance(old_par, bool):
                new_par = getval_yes_no(par, default=old_par)
            elif isinstance(old_par, int):
                new_par = getval_int(par, default=old_par)
            else:
                new_par = getval_float(par, default=old_par)
            self.parameters[det][par] = new_par

    def _info_categories(self) -> Dict[str, dict]:
        categories = super()._info_categories()
        ntasks = sum([plotter.purge_tasks() for _, plotter in self._plotters.items()])
        status = categories.setdefault("status", {})
        status["Plotting tasks"] = ntasks
        return categories

    def _get_workflow(self) -> dict:
        with open(resource_filename("id32", self.WORKFLOW_FILENAME), "r") as wf:
            return json.load(wf)

    def _get_workflow_inputs(self, scan: BlissScanType, detector) -> list:
        return [
            {
                "name": "scan_number",
                "value": scan.scan_number,
            },
            {
                "name": "input_path",
                "value": scan.scan_saving.filename,
            },
            {
                "name": "output_path",
                "value": self._get_scan_processed_directory(scan),
            },
            {"name": "detector", "value": detector},
            {"name": "specgen_parameters", "value": dict(self.parameters[detector])},
            {"name": "update_period", "value": self.update_period},
            {
                "name": "scan_info",
                "value": scan.scan_info["acquisition_chain"]["timer"],
            },
        ]

    def _get_scan_filename(self, scan: BlissScanType) -> str:
        filename = scan.scan_info.get("filename")
        if filename:
            return filename
        return current_session.scan_saving.filename

    def _get_scan_processed_directory(self, scan: BlissScanType) -> str:
        return get_dataset_processed_dir(self._get_scan_filename(scan))

    def _get_workflow_destination(self, scan: BlissScanType) -> str:
        """Builds the path where the workflow JSON will be saved."""
        filename = self._get_scan_filename(scan)
        scan_nb = scan.scan_info.get("scan_nb")
        root = self._get_scan_processed_directory(scan)
        stem = os.path.splitext(os.path.basename(filename))[0]
        wf_path = os.path.join(root, "workflows")
        pathlib.Path(wf_path).mkdir(parents=True, exist_ok=True)
        basename = f"{stem}_{scan_nb:04d}.json"
        return os.path.join(wf_path, basename)

    def _trigger_workflow_on_new_scan(self, scan: BlissScanType) -> None:
        if not scan.scan_info["save"]:
            return

        if "timer" not in scan.scan_info["acquisition_chain"].keys():
            logger.warning(
                "Not a timescan or loopscan. Ignoring specgen workflow trigger. "
                f"Disabling {str(self.detectors)} in measurement group will speed up acquisition."
            )
            return

        for detector in self.detectors:
            if "%s:image" % detector in scan.scan_info["channels"].keys():
                workflow = self._get_workflow()
                inputs = self._get_workflow_inputs(scan, detector)
                kwargs = {"inputs": inputs, "outputs": [{"all": False}]}
                kwargs["convert_destination"] = self._get_workflow_destination(scan)

                future = submit(args=(workflow,), kwargs=kwargs, queue=self.QUEUE)

                # data = future.result()
                datasetname = "_".join(
                    [
                        scan.scan_saving.collection_name,
                        scan.scan_saving.dataset_name,
                    ]
                )

                scan_number = scan.scan_number
                output_path = self._get_scan_processed_directory(scan)
                outputfile_sum = os.path.join(
                    output_path,
                    # ~ detector,
                    f"Online_analysis_{datasetname}_{detector}.spec",
                )
                self._plotters[detector].handle_workflow_result(
                    future, outputfile_sum, scan_number
                )


class Plotter(WithFlintAccess):
    def __init__(self, max_scans: int, update_period: float, detector: str) -> None:
        super().__init__()
        self._tasks = list()
        self.max_scans = max_scans
        self.update_period = max(0.5, update_period)
        self.detector = detector

    def purge_tasks(self) -> int:
        """Remove references to tasks that have finished."""
        self._tasks = [t for t in self._tasks if t]
        return len(self._tasks)

    def kill_tasks(self) -> int:
        """Kill all tasks."""
        gevent.killall(self._tasks)
        return self.purge_tasks()

    def handle_workflow_result(self, *args, **kwargs) -> None:
        """Handle workflow results in a background task"""
        self._spawn(self._handle_workflow_result, *args, **kwargs)

    def _handle_workflow_result(self, future, outputfile_sum: str, scan_number: int):
        try:
            while not future.done():
                self._update_plot(outputfile_sum, scan_number)
                time.sleep(self.update_period)
            self._update_plot(outputfile_sum, scan_number)
        except Exception as e:
            logger.exception(e)

    def _spawn(self, *args, **kw):
        task = gevent.spawn(*args, **kw)
        self._tasks.append(task)
        self.purge_tasks()

    def _get_plot(self) -> SpecGenPlot:
        return super()._get_plot("SpecGen (%s)" % (self.detector), SpecGenPlot)

    def _update_plot(self, outputfile_sum, scan_number):
        plot = self._get_plot()
        plot.update_plot(outputfile_sum, scan_number, self.max_scans)



import os
from typing import Any
from typing import Dict
from typing import Optional

from ...id12.converter import Id12Hdf5ToAsciiConverter
from .. import EWOKS_RESULTS_DIR


class DemoId12Hdf5ToSpecConverter(Id12Hdf5ToAsciiConverter):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        root_dir = os.path.join(EWOKS_RESULTS_DIR, "id12", "inhouse")
        defaults.setdefault(
            "external_proposal_outdir", os.path.join(root_dir, "EXTERNAL")
        )
        defaults.setdefault(
            "inhouse_proposal_outdir", os.path.join(root_dir, "INHOUSE2")
        )
        defaults.setdefault("test_proposal_outdir", os.path.join(root_dir, "NOBACKUP"))

        super().__init__(config=config, defaults=defaults)



from __future__ import annotations

from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ewoksjob.client import submit
from ewoksutils.task_utils import task_inputs

from ..import_utils import unavailable_function
from ..persistent.parameters import ParameterInfo
from ..processor import BaseProcessorWithPlotting
from ..processor import BlissScanType
from ..utils.directories import get_filename
from ..utils.directories import master_output_filename
from ..utils.directories import workflow_destination
from ..xrpd.plotter import XrpdPlotter

try:
    from id09.status import get_xray_energy
except ImportError as ex:
    get_xray_energy = unavailable_function(ex)


class TxsProcessor(
    BaseProcessorWithPlotting,
    parameters=[
        ParameterInfo("distance", category="Txs", doc="meter", validator=float),
        ParameterInfo("center", category="Txs", doc="pixel (hor, ver)"),
        ParameterInfo("binning", category="Txs", doc="(hor, ver)"),
        ParameterInfo("detector", category="Txs", validator=str),
        ParameterInfo("pixel", category="Txs", doc="meter (hor, ver)"),
        ParameterInfo("energy", category="Txs", doc="eV", validator=float),
        ParameterInfo("integrate1d_options", category="Txs"),
    ],
):
    plotter_class = XrpdPlotter

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("distance", 1)
        defaults.setdefault("center", (960, 960))
        defaults.setdefault("binning", 2)
        defaults.setdefault("detector", "rayonix")
        defaults.setdefault("pixel", None)
        defaults.setdefault("energy", "automatic")
        defaults.setdefault("integrate1d_options", dict())

        super().__init__(config=config, defaults=defaults)

    def _trigger_workflow_on_new_scan(self, scan: BlissScanType) -> None:
        self.trigger_workflow_on_new_scan(scan)

    def trigger_workflow_on_new_scan(self, scan: BlissScanType) -> None:
        if not self.scan_requires_processing(scan):
            return None

        workflow = self.get_workflow(scan)
        kwargs = {"inputs": self.get_inputs(scan), "outputs": [{"all": False}]}
        if scan.scan_info.get("save"):
            kwargs["convert_destination"] = workflow_destination(scan)

        future = submit(args=(workflow,), kwargs=kwargs, queue="celery")

        if self._plotter:
            scan_nb = scan.scan_info.get("scan_nb")
            self._plotter.handle_workflow_result(
                future,
                self.detector,
                scan_name=f"{scan.name}: {scan_nb}.1 {scan.name}",
                output_url=(
                    # TODO: Do not hard-code the output url
                    f"{master_output_filename(scan)}::/{scan_nb}.1/integrate/integrated"
                    if scan.scan_info.get("save")
                    else None
                ),
            )

    def _get_txs_task_identifier(self, scan: BlissScanType):
        if scan.scan_info.get("save"):
            return "ewokstxs.tasks.txs.TxsTask"
        return "ewokstxs.tasks.txs.TxsTaskWithoutSaving"

    def get_workflow(self, scan: BlissScanType):
        return {
            "graph": {"id": "txs"},
            "nodes": [
                {
                    "id": "txs_task",
                    "task_type": "class",
                    "task_identifier": self._get_txs_task_identifier(scan),
                }
            ],
        }

    def scan_requires_processing(self, scan: BlissScanType) -> bool:
        return f"{self.detector}:image" in scan.scan_info.get("channels", dict())

    def get_inputs(self, scan: BlissScanType) -> List[dict]:
        if self.energy == "automatic":
            energy = get_xray_energy()
        else:
            energy = self.energy

        return task_inputs(
            task_identifier=self._get_txs_task_identifier(scan),
            inputs={
                "scan_key": scan._scan_data.key,
                "filename": get_filename(scan),
                "scan": scan.scan_info.get("scan_nb"),
                "energy": energy,
                "distance": self.distance,
                "center": self.center,
                "detector": self.detector,
                "pixel": self.pixel,
                "binning": (self.binning, self.binning),
                "output_filename": master_output_filename(scan),
                "integrate1d_options": self.integrate1d_options.to_dict(),
            },
        )



from typing import Any
from typing import Dict
from typing import Optional

from ..bliss_globals import setup_globals
from ..import_utils import UnavailableObject
from ..import_utils import is_available
from ..streamline.scanner import StreamlineScanner

try:
    from streamline_changer.sample_changer import SampleChanger
except ImportError as ex:
    streamline_changer = UnavailableObject(ex)


class StreamlineSesScanner(StreamlineScanner):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ):
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("sample_changer_name", "streamline_sc")

        super().__init__(config=config, defaults=defaults)

    def measure_sample(self, *args, has_qrcode: bool = True, **kwargs):
        return None

    def _get_calibration(self):
        return {"non_empty": None}

    def _newsample(self, sample_name: str):
        print("NEW SAMPLE:", sample_name)


if is_available(setup_globals):
    streamline_sc = SampleChanger(
        setup_globals.streamline_translation, setup_globals.streamline_wago
    )
else:
    streamline_sc = UnavailableObject(ImportError)

streamline_scanner = StreamlineSesScanner()



from typing import Any
from typing import Dict
from typing import Optional

from ...bliss_globals import setup_globals
from ...exafs.processor import ExafsProcessor
from ...resources import resource_filename


class DemoExafsProcessor(ExafsProcessor):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("workflow", resource_filename("exafs", "exafs.ows"))
        defaults.setdefault("_scan_type", "any")
        counters = defaults.setdefault("_counters", dict())

        if self._HAS_BLISS:
            energy_unit = setup_globals.energy.unit or "eV"
            counters.setdefault(
                "any",
                {
                    "mu_name": "mu",
                    "energy_name": "energy",
                    "energy_unit": energy_unit,
                },
            )

        super().__init__(config=config, defaults=defaults)

    def _scan_type_from_scan(self, scan) -> Optional[str]:
        return "any"

    def _multi_xas_scan(self, scan) -> bool:
        # Single scan are also considered "multi-scan" so the
        # data is split (eventhough there is nothing to split,
        # we will only have one scan) and saved in PROCESSED_DATA.
        return True

    def _multi_xas_subscan_size(self, scan) -> int:
        # The workflow will look an monotonicity of the energy
        # counter to figure this out.
        return 0



from ..flint.access import WithFlintAccess


class Id02Plotter(WithFlintAccess):
    def __init__(self, number_of_scans: int) -> None:
        super().__init__()
        self._number_of_scans = number_of_scans

    def replot(self):
        pass



import os
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from .fluoxas import fluoxas_paths
from .fluoxas import fluoxas_workflow_inputs
from .mosaic_xrfmap import mosaic_xrfmap_paths
from .mosaic_xrfmap import mosaic_xrfmap_workflow_inputs
from .utils.execute import execute_graph
from .xrfmap import xrfmap_paths
from .xrfmap import xrfmap_workflow_inputs


def xrfmap(*args, **kargs):
    workflow_filename, inputs, convert_destination = xrfmap_parameters(*args, **kargs)
    execute_graph(
        workflow_filename, inputs=inputs, convert_destination=convert_destination
    )


def mosaic_xrfmap(*args, **kargs):
    workflow_filename, inputs, convert_destination = mosaic_xrfmap_parameters(
        *args, **kargs
    )
    execute_graph(
        workflow_filename, inputs=inputs, convert_destination=convert_destination
    )


def fluoxas(*args, **kargs):
    workflow_filename, inputs, convert_destination = fluoxas_parameters(*args, **kargs)
    execute_graph(
        workflow_filename, inputs=inputs, convert_destination=convert_destination
    )


def xrfmap_parameters(
    session: str,
    sample: str,
    dataset: str,
    scan_number: int,
    config_filenames: Sequence[str],
    detector_numbers: Sequence[int],
    instrument_name: str = None,
    fast_fitting: bool = True,
    quantification: bool = True,
    diagnostics: bool = True,
    livetime_ref_value: Union[str, int, float, None] = None,
    counter_ref_value: Union[str, int, float, None] = None,
    counter_name: Optional[str] = None,
    energy_name: Optional[str] = None,
    dirname: Optional[str] = None,
    resolution: Optional[Dict[str, Tuple[Union[int, float], str]]] = None,
    demo: bool = False,
) -> Tuple[str, List[dict], str]:
    paths = xrfmap_paths(
        session,
        sample,
        dataset,
        scan_number,
        config_filenames,
        dirname=dirname,
        demo=demo,
    )

    workflow, inputs = xrfmap_workflow_inputs(
        paths.filename,
        paths.output_root_uri,
        scan_number,
        paths.config_filenames,
        instrument_name=instrument_name,
        detector_numbers=detector_numbers,
        detector_names=None,
        fast_fitting=fast_fitting,
        quantification=quantification,
        diagnostics=diagnostics,
        livetime_ref_value=livetime_ref_value,
        counter_ref_value=counter_ref_value,
        counter_name=counter_name,
        energy_name=energy_name,
        resolution=resolution,
    )

    workflow_filename = os.path.join(paths.workflow_path, workflow + ".ows")

    return workflow_filename, inputs, paths.convert_destination


def mosaic_xrfmap_parameters(
    session: str,
    sample: str,
    datasets: Sequence[str],
    scan_ranges: Sequence[Tuple[int, int]],
    config_filenames: Sequence[str],
    detector_numbers: Sequence[int],
    instrument_name: str = None,
    fast_fitting: bool = True,
    quantification: bool = True,
    diagnostics: bool = True,
    livetime_ref_value: Union[str, int, float, None] = None,
    counter_ref_value: Union[str, int, float, None] = None,
    counter_name: Optional[str] = None,
    energy_name: Optional[str] = None,
    dirname: Optional[str] = None,
    exclude_scans: Optional[Sequence[Sequence[int]]] = None,
    resolution: Optional[Dict[str, Tuple[Union[int, float], str]]] = None,
    demo: bool = False,
):
    paths = mosaic_xrfmap_paths(
        session,
        sample,
        datasets,
        scan_ranges,
        config_filenames,
        dirname=dirname,
        demo=demo,
    )

    workflow, inputs = mosaic_xrfmap_workflow_inputs(
        paths.filenames,
        paths.output_root_uri,
        paths.concat_bliss_scan_uri,
        scan_ranges,
        paths.config_filenames,
        instrument_name=instrument_name,
        detector_numbers=detector_numbers,
        detector_names=None,
        fast_fitting=fast_fitting,
        quantification=quantification,
        diagnostics=diagnostics,
        livetime_ref_value=livetime_ref_value,
        counter_ref_value=counter_ref_value,
        counter_name=counter_name,
        energy_name=energy_name,
        exclude_scans=exclude_scans,
        resolution=resolution,
    )

    workflow_filename = os.path.join(paths.workflow_path, workflow + ".ows")

    return workflow_filename, inputs, paths.convert_destination


def fluoxas_parameters(
    session: str,
    sample: str,
    datasets: Sequence[str],
    scan_ranges: Sequence[Tuple[int, int]],
    config_filenames: Sequence[str],
    detector_numbers: Sequence[int],
    instrument_name: str = None,
    fast_fitting: bool = True,
    quantification: bool = True,
    diagnostics: bool = True,
    livetime_ref_value: Union[str, int, float, None] = None,
    counter_ref_value: Union[str, int, float, None] = None,
    counter_name: Optional[str] = None,
    energy_name: Optional[str] = None,
    dirname: Optional[str] = None,
    exclude_scans: Optional[Sequence[Sequence[int]]] = None,
    resolution: Optional[Dict[str, Tuple[Union[int, float], str]]] = None,
    do_align: bool = True,
    align_counter: Optional[str] = None,
    align_crop: Optional[bool] = True,
    fast_align_counter_selection: Optional[bool] = False,
    demo: bool = False,
):
    paths = fluoxas_paths(
        session,
        sample,
        datasets,
        scan_ranges,
        config_filenames,
        dirname=dirname,
        demo=demo,
    )

    workflow, inputs = fluoxas_workflow_inputs(
        paths.filenames,
        paths.output_root_uri,
        scan_ranges,
        paths.config_filenames,
        instrument_name=instrument_name,
        detector_numbers=detector_numbers,
        detector_names=None,
        fast_fitting=fast_fitting,
        quantification=quantification,
        diagnostics=diagnostics,
        livetime_ref_value=livetime_ref_value,
        counter_ref_value=counter_ref_value,
        counter_name=counter_name,
        energy_name=energy_name,
        exclude_scans=exclude_scans,
        resolution=resolution,
        do_align=do_align,
        align_counter=align_counter,
        align_crop=align_crop,
        fast_align_counter_selection=fast_align_counter_selection,
    )

    workflow_filename = os.path.join(paths.workflow_path, workflow + ".ows")

    return workflow_filename, inputs, paths.convert_destination



from importlib.metadata import version
from typing import Any
from typing import Dict

from packaging.version import Version

DEFINITIONS_VERSION = Version(version("icat-esrf-definitions"))
SUPPORTS_TECHNIQUE_PID = DEFINITIONS_VERSION >= Version("2.0.0")


def adapt_legacy_metadata(metadata: Dict[str, Any]) -> None:
    """Modifies ``metadata`` in case the installed ``icat-esrf-definitions``
    library does not support the latest ICAT metadata schema.
    """
    if not SUPPORTS_TECHNIQUE_PID:
        _ = metadata.pop("technique_pid")
